"""
Title: XmlConstructor

Author: PhanterJR<junior.conex@gmail.com>

License: MIT

Coding: utf-8

Tools that can be used in PhanterPWA.
"""

import re
import sys
import os
import json
import time
import shutil
from datetime import datetime, date
from pathlib import PurePath
from subprocess import Popen, PIPE, call
from unicodedata import normalize
from phanterpwa import __install_requeriments__ as requeriments
from tornado import escape


def check_requeriments() -> "generator - each step returns (str, bool)":
    """Each generator step returns a tuple with the name of the checked requirement and a boolean, True for Passed and
        False for Failed
    """
    for x in requeriments:
        if ">" in x or "<" in x or "=" in x:
            x = x.split(">")[0]
            x = x.split("=")[0]
            x = x.split("<")[0]
        checked = call(["pip", "show", x])
        yield (x, not bool(checked))


def search_google_chrome():
    if sys.platform == "win32":
        winvars = [
            '%PROGRAMFILES(X86)%\\Google',
            '%PROGRAMW6432%\\Google',
            '%PROGRAMFILES%\\Google',
            '%LOCALAPPDATA%\\Google',
        ]
        new_winvars = []
        for x in winvars:
            p = Popen('echo "{0}"'.format(x), stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
            ret, error = p.communicate()
            if ret.strip():
                if x not in new_winvars:
                    new_winvars.append(x)
        for x in new_winvars:
            p = Popen('where /r "{0}" "chrome.exe"'.format(x), stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
            ret, error = p.communicate()
            if ret.strip():
                return ret.strip()

    elif sys.platform.startswith('linux'):
        names = [
            'google-chrome',
            'google-chrome-stable'
            'chromium-browser',
            'chromium',
        ]
        for x in names:
            p = Popen('which {0}'.format(x), stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
            ret, error = p.communicate()
            if ret.strip():
                return ret.strip()
    else:
        return None


def interpolate(xstring, context: dict, delimiters: (list, tuple)=["{{", "}}"], ignore_non_strings: bool=False):
    """Interpolates by replacing strings in certain delimiters

    :param xstring: String that will be interpolated
    :param context: Dictionary containing the strings that will be replaced, the keys will be tracked in the string and
        the values will be replaced by the match found.
    :param delimiters: Is the pair of strings that delimit each variable contained in xstring
    :param ignore_non_strings: By default, all xstring will be converted to str, if set to True, different string
        values or XmlConstructor will not be changed.

    Example:
        >>> from phanterpwa.tools import interpolate
        >>> r = interpolate(
        ...     "long string: Ford {{model1}}, Chevrolet {{model2}}.",
        ...     {"model1": "Focus", "model2": "Celta"}
        ... )
        >>> print(r)
        long string: Ford Focus, Chevrolet Celta.
    """

    if not (isinstance(delimiters, (list, tuple)) and len(delimiters) == 2):
        raise ValueError("The delimiters must be list or tuple type and need two values.")
    if ignore_non_strings:
        from phanterpwa.xmlconstructor import XmlConstructor
        if not isinstance(xstring, (str, XmlConstructor)):
            return xstring
    xstring = str(xstring)
    if isinstance(context, dict):
        for x in context:
            k = "".join([delimiters[0], x, delimiters[1]])
            ls = xstring.split(k)
            ns = str(context[x]).join(ls)
            xstring = ns
        return xstring
    else:
        raise ValueError("The context from _format method must be dict. Given: {0}".format(type(context)))


def sass_change_vars(string_sass: str, context: dict={}) -> str:
    """This function is used to change the values of variables in a sass file

    :param string_sass: String with the sass code
    :param context: The dictionary keys in context represent the variables that will be searched for in the sass code
        with their respective new values.

    Example:
        >>> from phanterpwa.tools import sass_change_vars
        >>> sass_str = '''
        ... $BG: red
        ... $FG: green
        ... .my_class
        ...     background-color: $BG
        ...     color: $FG
        ... '''
        >>> print(sass_change_vars(sass_str, {"BG": blue, "FG": "black"}))
        $BG: blue
        $FG: black
        .my_class
            background-color: $BG
            color: $FG
    """

    ns = ""
    if all([isinstance(context, dict), isinstance(string_sass, str), context, string_sass]):
        lines = string_sass.split('\n')
        for x in lines:
            find = False
            for y in context:
                v = "".join(["$", y, ":"])
                if v in x:
                    ini = x.index(v)
                    ns = "".join([ns, x[:ini], v, " ", context[y], "\n"])
                    find = True
                    break
            if not find:
                ns = "".join([ns, x, "\n"])
        return ns[:-1]

    elif not isinstance(context, dict):
        raise ValueError("The context must be dict type. Given: {0}".format(type(context)))

    elif not isinstance(string_sass, str):
        raise ValueError("The string_sass must be str type. Given: {0}".format(type(string_sass)))

    elif not string_sass:
        raise ValueError("The string_sass is invalid. Given: {0}".format(string_sass))

    return ""


def sass_map_vars(string_sass):
    """Returns a dictionary with all the variables found with their respective values.

    :param string_sass: String with the sass code

    Example:
        >>> from phanterpwa.tools import sass_map_vars
        >>> sass_str = '''
        ... $BG: red
        ... $FG: green
        ... .my_class
        ...     background-color: $BG
        ...     color: $FG
        ... '''
        >>> sass_map_vars(sass_str)
        {'BG': 'red', 'FG': 'green'}
    """

    v = re.compile(r"^[\t ]*?(\$[A-Za-z_][A-Za-z0-9_\-]*\:{1}[ ]?[\w!#$%,\ .\-\(\)]+)\/*.*$", re.MULTILINE)
    r = v.findall(string_sass)
    d = {}
    for x in r:
        k, v = x.split(":")
        d[k.strip()[1:]] = v.strip()
    return d


def temporary_password(
        size: int = 8, chars: str = "abcdefghijlmnopqrstuvxzwykABCDEFGHIJLMNOPQRSTUVXZWYK0123456789") -> str:
    """Create a random sequence of characters.

    :param size: generated string size.
    :param chars: Characters that will be used to make the new string.

    Example:
        >>> from phanterpwa.tools import temporary_password
        >>> temporary_password()
        'j2LGe97s'
        >>> temporary_password(5)
        'OkPw7'
        >>> temporary_password(10, 'aeiou')
        'eioueauaiu'
    """

    if not (isinstance(size, int) and size > 0):
        raise ValueError("The size must be integer type and greater than 0")

    if not (isinstance(chars, str) and len(chars) > 2):
        raise ValueError("The chars must be string type and be at least 3 characters")

    import random
    len_chars = len(chars)
    matrix = chars
    p = []
    while len(p) < size:
        number = random.randint(0, len(matrix) - 1)
        char = matrix[number]
        if size <= len_chars:
            if char not in p:
                p.append(char)
        else:
            if p:
                if p[-1] != char:
                    p.append(char)
            else:
                p.append(char)
    f = "".join(p)
    return f


def app_name_from_relative_child(project_path, child_path):
    """Returns the application's folder name via a child folder

    :param project_path: Project path
    :child_path: Child folder of the project folder
    """
    apps_list_basedir = os.path.join(project_path, "frontapps")
    p = PurePath(child_path)
    r = p.relative_to(os.path.join(apps_list_basedir))
    return r.parts[0]


def generate_activation_code(size: int = 6) -> str:
    """Create a numeric code of predefined size with one character validator.

    :param size: generated string number size.

    Example:
        >>> from phanterpwa.tools import generate_activation_code
        >>> generate_activation_code()
        '425938-E'
        >>> generate_activation_code()
        '762093-A'
        >>> generate_activation_code(4)
        '3624-G'
    """

    if not (isinstance(size, int) and size > 0):
        raise ValueError("The size must be an integer and greater than 0")

    import random
    matrix = '0123456789'
    ver = "ABCDEFGHI"
    p = []
    su = 0
    while len(p) < size:
        number = random.randint(0, len(matrix) - 1)
        char = matrix[number]
        if char not in p:
            p.append(char)
            su += int(char)
            if su >= 9:
                su = su - 9
    f = "".join(p)
    return "-".join([f, ver[su]])


def check_activation_code(code: str, size: int = 6) -> (str, None):
    """Checks whether the string in the code parameter is a valid activation code

    :param size: Activation code size ignoring the check digit

    Example:
        >>> from phanterpwa.tools import (generate_activation_code, check_activation_code)
        >>> generate_activation_code()
        '425938-E'
        >>> check_activation_code('425938-E')
        '425938-E'
        >>> check_activation_code('INVALID_CODE')
        >>> print(check_activation_code('INVALID_CODE'))
        None
        >>> generate_activation_code(4)
        '3624-G'
        >>> check_activation_code('3624-G')
        >>> print(check_activation_code('3624-G'))
        None
        >>> check_activation_code('3624-G', 4)
        '3624-G'
    """

    if not (isinstance(size, int) and size > 0):
        raise ValueError("The size must be an integer and greater than 0")

    if not (isinstance(code, str) and len(code) > 2):
        raise ValueError("The chars must be an integer and be at least 3 characters")

    code = code.strip()
    if (len(code) == size + 2) and "-" in code:
        cod, dig = code.split("-")
        ver = "ABCDEFGHI"
        su = 0
        for char in cod:
            su += int(char)
            if su >= 9:
                su = su - 9
        if ver[su] == dig:
            return code


def config(cfg_file, dict_cfg=None, rewrite=False):
    """With this function it is possible to manipulate configuration files in json format, by default the file name is
    config.json

    :param cfg_file: path to the configuration file
    :param dict_cfg: dictionary with the values you want to add or change.
    :param rewrite: With this option the current file will be overwritten with the new settings added in the dict_cfg
        parameter

    Example:
        >>> from phanterpwa.tools import config
        >>> config("config.json", {"new_key": None})
        {'new_key': None}
        >>> with open("config.json") as f:
        ...     print(f.read())
        ...
        {
            "new_key": null
        }
        >>> config("config.json", {"new_key": False, "other_key": True})
        {'new_key': False, 'other_key': True}
        >>> with open("config.json") as f:
        ...     print(f.read())
        ...
        {
            "new_key": null,
            "other_key": true
        }
        >>> config("config.json", {"more_one_key": "Is it cumulative?"})
        {'new_key': False, 'other_key': True, 'more_one_key': "Is it cumulative?"}
        >>> with open("config.json") as f:
        ...     print(f.read())
        ...
        {
            "new_key": null,
            "other_key": true,
            "more_one_key": "Is it cumulative?"
        }
        >>> config("config.json", {"rewrite": "yes"}, True)
        {'rewrite': 'yes'}
        >>> with open("config.json") as f:
        ...     print(f.read())
        ...
        {
            "rewrite": "yes"
        }

    """
    if dict_cfg is None:
        dict_cfg = dict()

    if not isinstance(dict_cfg, dict):
        ValueError("The dict_cfg must be sict type. Given: {0}".format(type(dict_cfg)))

    if isinstance(cfg_file, str):
        if len(cfg_file) > 5:
            if cfg_file[-5:] != ".json":
                if os.path.exists(cfg_file) and os.path.isdir(cfg_file):
                    cfg_file = os.path.join(cfg_file, "config.json")
                else:
                    cfg_file = "".join([cfg_file, ".json"])
        else:
            if os.path.exists(cfg_file) and os.path.isdir(cfg_file):
                cfg_file = os.path.join(cfg_file, "config.json")
            else:
                cfg_file = "".join(cfg_file, ".json")
    else:
        raise ValueError("The arg 'cfg_file' must be strig. Given: {0}".format(type(cfg_file)))
    if os.path.exists(cfg_file) and not rewrite:
        cfg = None
        with open(cfg_file, 'r', encoding="utf-8") as f:
            cfg = json.load(f)
        if cfg and isinstance(cfg, dict):
            for x in dict_cfg:
                cfg[x] = dict_cfg[x]
        else:
            cfg = dict_cfg
    else:
        basedir = os.path.dirname(cfg_file)
        cfg = dict_cfg
        if basedir:
            os.makedirs(basedir, exist_ok=True)
    ncfg = dict(cfg)
    with open(cfg_file, "w", encoding="utf-8") as f:
        json.dump(ncfg, f, ensure_ascii=True, indent=2)
    return ncfg


def split_seconds(seconds: int) -> dict:
    """This function converts seconds into a dictionary by splitting seconds without year, month, day, hour, minute
    and second when possible.

    :param seconds: seconds that will be converted

    Example:
        >>> from phanterpwa.tools import split_seconds
        >>> split_seconds(123456789)
        {'year': 3, 'month': 11, 'day': 3, 'hour': 21, 'minute': 33, 'second': 9}
        >>> split_seconds(121)
        {'minute': 2, 'second': 1}
        >>> split_seconds(3659)
        {'hour': 1, 'second': 59}
    """

    if not isinstance(seconds, int):
        raise ValueError("The seconds must be an integer. Given: {0}".format(seconds))

    def s(seconds, d={}):
        d = dict(**d)
        if seconds >= 31536000:
            if seconds % 31536000:
                r = seconds % 31536000
                d['year'] = seconds // 31536000
                return s(r, d)
            else:
                d['year'] = seconds // 31536000
                return d
        else:
            if seconds >= 2592000:
                if seconds % 2592000:
                    r = seconds % 2592000
                    d['month'] = seconds // 2592000
                    return s(r, d)
                else:
                    d['month'] = seconds // 2592000
                    return d
            else:
                if seconds >= 86400:
                    if seconds % 86400:
                        r = seconds % 86400
                        d['day'] = seconds // 86400
                        return s(r, d)
                    else:
                        d['day'] = seconds // 86400
                        return d
                else:
                    if seconds >= 3600:
                        if seconds % 3600:
                            r = seconds % 3600
                            d['hour'] = seconds // 3600
                            return s(r, d)
                        else:
                            d['hour'] = seconds // 3600
                            return d
                    else:
                        if seconds >= 60:
                            if seconds % 60:
                                r = seconds % 60
                                d['minute'] = seconds // 60
                                return s(r, d)
                            else:
                                d['minute'] = seconds // 60
                                return d
                        else:
                            d['second'] = seconds
                            return d
    return s(seconds)


def join_seconds(splitted_seconds: dict) -> int:
    """Different from split_seconds, this function transforms a dictionary with the year, month, day, hour, minute and
    second keys with their respective values in seconds.
    
    :param splitted_seconds: dictionary with the year, month, day, hour, minute or
        second keys.

    Example:
        >>> from phanterpwa.tools import join_seconds
        >>> join_seconds({'year': 1, 'hour': 1, 'second': 1})
        31539601
        >>> join_seconds({'year': 1, 'second': 3601})
        31539601
        >>> join_seconds({'year': 1, 'minute': 60, 'second': 1})
        31539601
    """

    if isinstance(splitted_seconds, dict):
        total = 0
        if "year" in splitted_seconds:
            total += splitted_seconds["year"] * 60 * 60 * 24 * 365
        if "month" in splitted_seconds:
            total += splitted_seconds["month"] * 60 * 60 * 24 * 30
        if "day" in splitted_seconds:
            total += splitted_seconds['day'] * 60 * 60 * 24
        if "hour" in splitted_seconds:
            total += splitted_seconds['hour'] * 60 * 60
        if "minute" in splitted_seconds:
            total += splitted_seconds['minute'] * 60
        if "second" in splitted_seconds:
            total += splitted_seconds['second']
        return total
    else:
        raise ValueError("The splitted_seconds must be dict type. Given: {0}".format(type(splitted_seconds)))


def humanize_seconds(seconds: int, translator_instance: ("phanterpwa.i18n.Translator", None)=None) -> str:
    """Humanizes seconds by dividing between year, month, day, hour, minute and second when possible.

    :param seconds: seconds that will be humanized
    :param translator_instance: With this option it is possible to add an i18n instance to translate the return.

    Example:
        >>> from phanterpwa.tools import humanize_seconds
        >>> humanize_seconds(123456789)
        '3 years, 11 months, 3 days, 21 hours, 33 minutes and 9 seconds'
        >>> humanize_seconds(121)
        '2 minutes and 1 second'
        {'minute': 2, 'second': 1}
        >>> humanize_seconds(3659)
        '1 hour and 59 seconds'

        Using phanterpwa.i18n.Translator example

        >>> from phanterpwa.i18n import Translator
        >>> my_translator = Translator("test_path", "humanize_seconds")
        >>> my_translator.translate("pt-BR", "year", "ano")
        {"pt-BR": {"year": "ano"}}
        >>> my_translator.translate("pt-BR", "years", "anos")
        {"pt-BR": {"years": "anos"}}
        >>> my_translator.translate("pt-BR", "month", "mês")
        {"pt-BR": {"month": "mês"}}
        >>> my_translator.translate("pt-BR", "months", "meses")
        {"pt-BR": {"months": "meses"}}
        >>> my_translator.translate("pt-BR", "day", "dia")
        {"pt-BR": {"day": "dia"}}
        >>> my_translator.translate("pt-BR", "days", "dias")
        {"pt-BR": {"days": "dias"}}
        >>> my_translator.translate("pt-BR", "hour", "hora")
        {"pt-BR": {"hour": "hora"}}
        >>> my_translator.translate("pt-BR", "hours", "horas")
        {"pt-BR": {"hours": "horas"}}
        >>> my_translator.translate("pt-BR", "minute", "minuto")
        {"pt-BR": {"minute": "minuto"}}
        >>> my_translator.translate("pt-BR", "minutes", "minutos")
        {"pt-BR": {"minutes": "minutos"}}
        >>> my_translator.translate("pt-BR", "second", "segundo")
        {"pt-BR": {"second": "segundo"}}
        >>> my_translator.translate("pt-BR", "seconds", "segundos")
        {"pt-BR": {"seconds": "segundos"}}
        >>> my_translator.direct_translation = "pt-BR"
        >>> humanize_seconds(123456789, my_translator)
        '3 anos, 11 meses, 3 dias, 21 horas, 33 minutos and 9 segundo'
        >>> my_translator.direct_translation = "en-US"
        >>> humanize_seconds(123456789, my_translator)
        '3 years, 11 months, 3 days, 21 hours, 33 minutes and 9 seconds'
        >>> humanize_seconds(3659, my_translator)
        '1 hour and 59 seconds'
        >>> my_translator.direct_translation = "pt-BR"
        >>> humanize_seconds(3659, my_translator)
        '1 hora and 59 segundos'
        >>> my_translator.translate("pt-BR", "and ", "e ")
        >>> humanize_seconds(3659, my_translator)
        '1 hora e 59 segundos'
    """

    if translator_instance:
        from phanterpwa.i18n import Translator
        if not isinstance(translator_instance, Translator):
            raise ValueError("The translator_instance must be instance of phanterpwa.i18n.Translator or None")
    if not isinstance(seconds, int):
        raise ValueError("The seconds must be an integer. Given: {0}".format(seconds))

    result = split_seconds(seconds)
    s = ""
    ts_list = []
    if "year" in result:
        kunit = unit = "year"
        if result[kunit] > 1:
            unit = "".join([unit, "s"])
        if translator_instance:
            unit = translator_instance.T(unit)
        ts = "".join([str(result[kunit]), " ", unit, " "])
        ts_list.append(ts)
    if "month" in result:
        kunit = unit = "month"
        if result[kunit] > 1:
            unit = "".join([unit, "s"])
        if translator_instance:
            unit = translator_instance.T(unit)
        ts = "".join([str(result[kunit]), " ", unit, " "])
        ts_list.append(ts)
    if "day" in result:
        kunit = unit = "day"
        if result[kunit] > 1:
            unit = "".join([unit, "s"])
        if translator_instance:
            unit = translator_instance.T(unit)
        ts = "".join([str(result[kunit]), " ", unit, " "])
        ts_list.append(ts)
    if "hour" in result:
        kunit = unit = "hour"
        if result[kunit] > 1:
            unit = "".join([unit, "s"])
        if translator_instance:
            unit = translator_instance.T(unit)
        ts = "".join([str(result[kunit]), " ", unit, " "])
        ts_list.append(ts)
    if "minute" in result:
        kunit = unit = "minute"
        if result[kunit] > 1:
            unit = "".join([unit, "s"])
        if translator_instance:
            unit = translator_instance.T(unit)
        ts = "".join([str(result[kunit]), " ", unit, " "])
        ts_list.append(ts)
    if "second" in result:
        kunit = unit = "second"
        if result[kunit] > 1:
            unit = "".join([unit, "s"])
        if translator_instance:
            unit = translator_instance.T(unit)
        ts = "".join([str(result[kunit]), " ", unit, " "])
        ts_list.append(ts)
    if len(ts_list) > 1:
        cont = 0
        for x in ts_list[:-1]:
            if cont:
                s = ", ".join([s.strip(), x])
            else:
                s = "".join([s, x])
            cont += 1
        else:
            if translator_instance:
                s = translator_instance.T("and ").join([s, ts_list[-1].strip()])
            else:
                s = "and ".join([s, ts_list[-1].strip()])
    else:
        if ts_list:
            s = ts_list[0].strip()

    return s


def text_normalize(text: str, upper: bool = True) -> str:
    """"Remove the graphic accents of Latin languages"

    :param text: string that will be normalized
    :param upper: This parameter with the value True (default) the output will be capitalized

    Example:
        >>> from phanterpwa.tools import text_normalize
        >>> text_normalize("Maçã")
        'MACA'
        >>> text_normalize("Maçã", False)
        'Maca'

    """

    if upper:
        return normalize('NFKD', text).encode('ascii', 'ignore').decode('ascii').upper()
    else:
        return normalize('NFKD', text).encode('ascii', 'ignore').decode('ascii')


def filter_for_a_valid_filename(value):
    texto = str(value)
    mini = "ÁÉÍÓÚÀÈÌÒÙÂÊÎÔÛÃÕÄËÏÖÜÇABCDEFGHIJKLMNOPQRSTUVWXYZ áéíóúàèìòùâêîôûãõäëïöüabçcdefghijlmnopqrstuvxzwyk0123456789ªº°[{()}]"
    novo_texto = ""
    for x in texto:
        if x in mini:
            novo_texto = "".join([novo_texto, x])
        else:
            novo_texto = "".join([novo_texto, " "])
    texto = one_space(novo_texto)
    return texto


def string_escape(value):
    r"""Escapes the <,>, ", 'and & characters of a string, when adding an interactable object such as a list,
    a dictionary, tuples and a set it cycles through the values in search of strings recursively, remembering
    that objects of the type tuple and set will be converted to lists.

    :param value: string, dict, list, tuple or set.

    Example:
        >>> from phanterpwa.tools import string_escape
        >>> string_escape("<div>the content<\div>")
        '&lt;div&gt;the content&lt;\div&gt;'
        >>> result = string_escape({
            "one": "<div style='test'>one</div>",
            "two": ["<div>two</div>"],
            "three": ("<div>three</div>", ),
            "four": {"<div>four</div>"},
            "five": {
                "five_one": ["<br>", ("<input>", "<hr>"), {"<a href='link'></a>"}]
            }
        })
        >>> result["one"]
        '&lt;div style=&#39;test&#39;&gt;one&lt;/div&gt;'
        >>> result["two"]
        '["&lt;div&gt;two&lt;/div&gt;"]'
        >>> result["three"]
        '["&lt;div&gt;three&lt;/div&gt;"]'
        >>> result["four"]
        '['&lt;div&gt;four&lt;/div&gt;']'
        >>> result["five"]["five_one"]
        '["&lt;br&gt;", ["&lt;input&gt;", "&lt;hr&gt;"], ["&lt;a href=&#39;link&#39;&gt;&lt;/a&gt;"]]'

    """
    if isinstance(value, str):
        return escape.xhtml_escape(value)
    elif isinstance(value, (list, tuple, set)):
        new_list = []
        for x in value:
            new_list.append(string_escape(x))
        return new_list
    elif isinstance(value, dict):
        new_dict = {}
        for x in value.keys():
            new_dict[x] = string_escape(value[x])
        return new_dict
    elif isinstance(value, datetime):
        return value.isoformat(" ", "seconds")
    elif isinstance(value, date):
        return value.isoformat()
    else:
        return value


def one_space(value):
    """"Removes empty spaces, tabs and new lines from a string and adds a space between words when necessary.

    Example:
        >>> from phanterpwa.tools import one_space
        >>> one_space("   My    long \r\n text. \tTabulation,      spaces,     spaces.         ")
        'My long text. Tabulation, spaces, spaces.'


    """
    result = ""
    if isinstance(value, str):
        value = value.strip()
        value = value.replace("\n", " ").replace("\t", " ").replace("\r", " ")
        spl = value.split(" ")
        result = " ".join([x for x in spl if x])
    return result


def user_agent_parse(user_agent):
    if not isinstance(user_agent, str):
        return "Unknown (Unknown)"
    keys =[
        "Edg",
        "CrOS",
        "Linux",
        "Edge",
        "Chromium",
        "OpenBSD",
        "Waterfox",
        "Lumia",
        "Trident",
        "Nexus",
        "Opera",
        "Android",
        "PaleMoon",
        "BingPreview",
        "Fedora",
        "Ubuntu",
        "Mac OS X",
        "Firefox",
        "iPhone",
        "IEMobile",
        "Tablet",
        "Safari",
        "Instagram",
        "iPad",
        "MSIE",
        "X11",
        "SeaMonkey",
        "Iceweasel",
        "SAMSUNG",
        "Win64",
        "Windows",
        "Puffin",
        "Maxthon",
        "Chrome",
        "MacX11",
        "PhantomJS",
        "Mobile",
        "OPR",
        "PlayStation",
        "WOW64",
        "UCBrowser",
        "FB",
        "iPhone OS",
        "YaBrowser",
        "Redmi",
        "Konqueror",
        "Macintosh"
    ]
    tagx = set(keys)


    for_print = list()
    for y in tagx:
        if y.upper() in user_agent.upper():
            if y == "Trident" or y == "MSIE":
                if "MSIE" not in for_print:
                    for_print.append("MSIE")
            elif y == "WOW64" or y == "Win64" or y == "Windows":
                if "Windows" not in for_print:
                    for_print.append("Windows")
            else:
                for_print.append(y.strip())
    if for_print:
        change = False
        if "Windows" in for_print:
            for_print.pop(for_print.index("Windows"))
            for_print.append("(Windows)")
        if "iPhone OS" in for_print:
            if "iPhone OS" in for_print:
                for_print.pop(for_print.index("iPhone OS"))
            if "Mac OS X" in for_print:
                for_print.pop(for_print.index("Mac OS X"))
            if "Macintosh" in for_print:
                for_print.pop(for_print.index("Macintosh"))
            if "Safari" in for_print and "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            for_print.append("(iOS)")
        if "Mobile" in for_print:
            if "iPad" in for_print:
                for_print.pop(for_print.index("Mobile"))
                for_print.pop(for_print.index("iPad"))
                if "iPhone" in for_print:
                    for_print.pop(for_print.index("iPhone"))
                for_print.insert(0, "iPad")
            if "iPhone" in for_print:
                for_print.pop(for_print.index("Mobile"))
                for_print.pop(for_print.index("iPhone"))
                for_print.insert(0, "iPhone")
            if "Android" in for_print:
                for_print.pop(for_print.index("Mobile"))
                for_print.pop(for_print.index("Android"))
                if "Safari" in for_print:
                    for_print.pop(for_print.index("Safari"))
                if "Linux" in for_print:
                    for_print.pop(for_print.index("Linux"))
                for_print.append("(Android)")
        if "Android" in for_print:
            for_print.pop(for_print.index("Android"))
            if "Chrome" in for_print and "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Linux" in for_print:
                for_print.pop(for_print.index("Linux"))
            for_print.append("(Android)")           
        if "Edg" in for_print or "Edge" in for_print:
            if "Edg" in for_print:
                for_print.pop(for_print.index("Edg"))
            if "Edge" in for_print:
                for_print.pop(for_print.index("Edge"))
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            for_print.insert(0, "Edge")         
        if "X11" in for_print and "Linux" in for_print:
            for_print.pop(for_print.index("X11"))
            for_print.pop(for_print.index("Linux"))
            for_print.append("(Linux)")
        if "X11" in for_print and "CrOS" in for_print:
            for_print.pop(for_print.index("X11"))
            for_print.pop(for_print.index("CrOS"))
            for_print.append("(ChromeOS)")
        if "X11" in for_print and "OpenBSD" in for_print:
            for_print.pop(for_print.index("X11"))
            for_print.pop(for_print.index("OpenBSD"))
            for_print.append("(OpenBSD)")

        if "Chromium" in for_print:
            change=True
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Chromium" in for_print:
                for_print.pop(for_print.index("Chromium"))
            for_print.insert(0, "Chromium")

        if "Macintosh" in for_print or "Mac OS X" in for_print:
            if "Mac OS X" in for_print:
                for_print.pop(for_print.index("Mac OS X"))
            if "Macintosh" in for_print:
                for_print.pop(for_print.index("Macintosh"))
            if "Safari" in for_print and "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            for_print.append("(Mac OS X)")
        if "Safari" in for_print and "Chrome" in for_print:
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
        if "OPR" in for_print or "Opera" in for_print:
            if "OPR" in for_print:
                for_print.pop(for_print.index("OPR"))
            if "Opera" in for_print:
                for_print.pop(for_print.index("Opera"))
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            for_print.insert(0, "Opera")
        if "Nexus" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "PhantomJS" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "Maxthon" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "Puffin" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "SeaMonkey" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "SAMSUNG" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "BingPreview" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "PaleMoon" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "Iceweasel" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "PhantomJS" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "UCBrowser" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "YaBrowser" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "Instagram" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "Waterfox" in for_print:
            if "Chrome" in for_print:
                for_print.pop(for_print.index("Chrome"))
            if "Safari" in for_print:
                for_print.pop(for_print.index("Safari"))
            if "Firefox" in for_print:
                for_print.pop(for_print.index("Firefox"))
        if "Lumia" in for_print:

            for_print.pop(for_print.index("Lumia"))
            if "Mobile" in for_print:
                for_print.pop(for_print.index("Mobile"))
            if "IEMobile" in for_print and "MSIE" in for_print:
                for_print.pop(for_print.index("IEMobile"))
            for_print.insert(0, "Lumia")
        if "Redmi" in for_print:
            for_print.pop(for_print.index("Redmi"))
            for_print.insert(0, "Redmi")
        if "Tablet" in for_print:
            if "Tablet" in for_print:
                for_print.pop(for_print.index("Tablet"))
            for_print.insert(0, "Tablet")
        if "Linux" in for_print:
            for_print.pop(for_print.index("Linux"))
            for_print.append("(Linux)")
        if "FB" in for_print:
            for_print[for_print.index("FB")] = "Facebook"
        return " ".join([z for z in for_print])
    else:
        return "Unknown (Unknown)"


def checkbox_bool(value):
    if str(value).upper() in ["TRUE", "ON"]:
        return True
    else:
        return False


def normalize_names(name, not_captilize=[]):
    return " ".join([
        x if x in not_captilize or len(x) < 3 else x.capitalize() for x in name.lower().strip().split(" ") if x])


class WatchingFiles():
    def __init__(self, path_monitoring, path_destiny, ignore_paths=["__pycache__"], exec_after=None):
        self.path_monitoring = path_monitoring
        self.path_destiny = path_destiny
        self.ignore_paths = ignore_paths
        self._dirs = dict()
        self._files = dict()
        self.exec_after = exec_after
        self._create_mtime_list()

    @property
    def path_monitoring(self):
        return self._path_monitoring

    @path_monitoring.setter
    def path_monitoring(self, value):
        if os.path.isdir(value):
            self._path_monitoring = value
        else:
            raise IOError("The path \"{0}\" not exists!".format(value))

    def _create_mtime_list(self):
        print("Checking files modified at source\n")
        self._dirs = dict()
        self._files = dict()
        destiny_dirs = []
        destiny_files = []

        for x in os.walk(self.path_monitoring):
            p = PurePath(x[0])
            if not set(p.parts).intersection(set(self.ignore_paths)):
                df = os.path.normpath(os.path.join(self.path_destiny, p.relative_to(self.path_monitoring)))
                self._dirs[x[0]] = [os.path.getmtime(x[0]), df]
                for y in x[2]:
                    f = os.path.normpath(os.path.join(x[0], y))
                    d = os.path.normpath(os.path.join(self.path_destiny, p.relative_to(self.path_monitoring), y))
                    self._files[
                        f
                    ] = [os.path.getmtime(f), d]
                    if os.path.isfile(d):
                        if os.path.getmtime(f) != os.path.getmtime(d):
                            print("Was Modify: {0}\nCoping {0} to {1}".format(f, d))
                            shutil.copy2(f, d)

        if os.path.isdir(self.path_destiny):
            for x in os.walk(self.path_destiny):
                p = PurePath(x[0])
                if not set(p.parts).intersection(set(self.ignore_paths)):
                    df = os.path.normpath(os.path.normpath(x[0]))
                    destiny_dirs.append(df)
                    for y in x[2]:
                        d = os.path.normpath(os.path.join(x[0], y))
                        destiny_files.append(d)
        diff_dir = set(destiny_dirs).difference(set([self._dirs[x][1] for x in self._dirs.keys()]))
        diff_fil = set(destiny_files).difference(set([self._files[x][1] for x in self._files.keys()]))
        if diff_dir:
            for tt in diff_dir:
                if os.path.isdir(tt):
                    print("Deleting {0}".format(tt))
                    try:
                        shutil.rmtree(tt)
                    except OSError as e:
                        print("Error: %s - %s." % (e.filename, e.strerror))
                        pass
                    while os.path.exists(tt):
                        pass
        if diff_fil:
            for ff in diff_fil:
                if os.path.isfile(ff):
                    print("Deleting {0}".format(ff))
                    try:
                        os.unlink(ff)
                    except OSError as e:
                        print("Error: %s - %s." % (e.filename, e.strerror))
                        pass
                    while os.path.exists(ff):
                        pass

    def monitoring(self):
        print("Starting WatchingFiles...")
        self._has_change = False
        while True:
            self.sincronize()

    def sincronize(self):
        dirs_delete = dict(self._dirs)
        files_delete = dict(self._files)
        change_local = True
        for x in os.walk(self.path_monitoring):
            p = PurePath(x[0])
            if not set(p.parts).intersection(set(self.ignore_paths)):
                df = os.path.normpath(os.path.join(self.path_destiny, p.relative_to(self.path_monitoring)))
                self._dirs[x[0]] = [os.path.getmtime(x[0]), df]
                if x[0] in dirs_delete:
                    del dirs_delete[x[0]]
                if not os.path.isdir(df):
                    os.makedirs(df, exist_ok=True)
                    change_local = True
                    print("Creating Destiny Dirs: {0}".format(df))
                for y in x[2]:
                    f = os.path.normpath(os.path.join(x[0], y))
                    d = os.path.normpath(os.path.join(self.path_destiny, p.relative_to(self.path_monitoring), y))
                    if f in files_delete:
                        del files_delete[f]
                    if f not in self._files:
                        self._files[
                            f
                        ] = [os.path.getmtime(f), d]
                        print("New File: {0}\nCoping {0} to {1}".format(f, d))
                        change_local = True
                        shutil.copy2(f, d)
                    else:
                        if self._files[f][0] != os.path.getmtime(f):
                            self._files[
                                f
                            ] = [os.path.getmtime(f), d]
                            print("Was Modify: {0}\nCoping {0} to {1}".format(f, d))
                            change_local = True
                            shutil.copy2(f, d)
                        elif not os.path.isfile(d):
                            change_local = True
                            shutil.copy2(f, d)
                            print("Not Found: {1}\nCoping {0} to {1}".format(f, d))
        if dirs_delete:
            print("Deleting dirs:", dirs_delete)

            for dd in tuple(dirs_delete.keys()):
                target = dirs_delete[dd][1]
                print(target)
                if os.path.isdir(target):
                    try:
                        shutil.rmtree(target)
                    except OSError as e:
                        print("Error: %s - %s." % (e.filename, e.strerror))
                        pass
                    while os.path.exists(target):
                        pass
                    del dirs_delete[dd]
                    del self._dirs[dd]
                else:
                    del dirs_delete[dd]
                    del self._dirs[dd]

        if files_delete:
            print("Deleting files:", files_delete)

            for ff in tuple(files_delete.keys()):
                target = files_delete[ff][1]
                print(target)
                if os.path.isfile(target):
                    try:
                        os.unlink(target)
                    except OSError as e:
                        print("Error: %s - %s." % (e.filename, e.strerror))
                        pass
                    while os.path.exists(target):
                        pass
                    del files_delete[ff]
                    del self._files[ff]
                else:
                    del files_delete[ff]
                    del self._files[ff]
        if change_local and callable(self.exec_after):
            self.exec_after(self)
        change_local = False
        time.sleep(1)



# new decorators
import asyncio
import json
from functools import wraps
from ..security import (
    Serialize,
    SignatureExpired,
    BadSignature,
    URLSafeSerializer,
)


def check_application():
    def decorator(f):
        @wraps(f)
        def check_application_decorator(self, *args, **kargs):
            if not hasattr(self, "phanterpwa_client_application_checked"):
                self.phanterpwa_client_application_checked = None
                project_name = self.projectConfig['PROJECT']['name']
                project_version = self.projectConfig['PROJECT']['version']
                self.phanterpwa_application = self.request.headers.get('phanterpwa-application')
                self.phanterpwa_application_version = self.request.headers.get('phanterpwa-application-version')
                if self.phanterpwa_application == project_name:
                    if self.phanterpwa_application_version != project_version:
                        msg = 'The client needs update ({0}).'
                        dict_response = {
                            'status': 'Conflict',
                            'code': 409,
                            'specification': 'client needs update',
                            'message': msg.format(project_version),
                            'i18n': {
                                'message': self.i18nTranslator.T(msg).format(project_version)
                                    if self.i18nTranslator else msg.format(project_version)
                            }
                        }
                        self.set_status(400)
                        return dict_response
                else:
                    msg = 'The client is not compatible.'
                    dict_response = {
                        'status': 'Bad Request',
                        'code': 400,
                        'message': msg,
                        'i18n': {
                            'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                        },
                        'project_path': self.projectConfig['PROJECT']['path']
                    }
                    self.set_status(400)
                    return dict_response
                self.phanterpwa_client_application_checked = {
                    'name': self.phanterpwa_application,
                    'version': self.phanterpwa_application_version
                }
                return f(self, *args, **kargs)
            else:
                return f(self, *args, **kargs)
        return check_application_decorator
    return decorator


def check_client_token(ignore_locked=True):
    def decorator(f):
        @wraps(f)
        @check_application()
        def check_client_token_decorator(self, *args, **kargs):
            if not hasattr(self, "phanterpwa_client_token_checked"):
                self.phanterpwa_current_client = None
                self.phanterpwa_client_token_checked = None
                self.phanterpwa_authorization_checked = None
                self.phanterpwa_client_token = self.request.headers.get('phanterpwa-client-token')
                self.phanterpwa_authorization = self.request.headers.get('phanterpwa-authorization')
                self.phanterpwa_user_agent = self.request.headers.get('User-Agent')
                if not self.phanterpwa_client_token:
                    msg = 'Client token is not in the header. "phanterpwa-client-token"'
                    dict_response = {
                        'status': 'Bad Request',
                        'code': 400,
                        'message': msg,
                        'i18n': {
                            'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                        }
                    }
                    self.set_status(400)
                    return dict_response
                t = Serialize(
                    self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                    self.projectConfig['BACKEND'][self.app_name]['default_time_client_token_expire']
                )
                q = self.DALDatabase(self.DALDatabase.client.token == self.phanterpwa_client_token).select().first()
                if q:
                    token_content = None
                    try:
                        token_content = t.loads(self.phanterpwa_client_token)
                    except BadSignature:
                        token_content = None
                    except SignatureExpired:
                        token_content = None
                    if token_content:
                        if "user_agent" in token_content and "id_client" in token_content:
                            if token_content['user_agent'] == self.phanterpwa_user_agent and\
                                int(token_content['id_client']) == q.id:
                                if self.phanterpwa_authorization:
                                    t_user = Serialize(
                                        self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                                        self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
                                    )
                                    token_content_user = None
                                    id_user = None
                                    try:
                                        token_content_user = t_user.loads(self.phanterpwa_authorization)
                                    except BadSignature:
                                        token_content_user = None
                                    except SignatureExpired:
                                        token_content_user = None
                                    if token_content_user and 'id' in token_content_user:
                                        self.phanterpwa_authorization_checked = token_content_user
                                        id_user = token_content_user['id']
                                    if id_user and 'id_user' in token_content and id_user == token_content['id_user']:
                                        self.phanterpwa_client_token_checked = token_content
                                else:
                                    self.phanterpwa_client_token_checked = token_content

                if self.phanterpwa_client_token_checked:
                    if q and q.locked and not ignore_locked:
                        msg = "The user has locked your session!"
                        dict_response = {
                            'status': 'Unauthorized',
                            'code': 401,
                            'specification': 'client locked',
                            'message': msg,
                            'i18n': {
                                'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                            }
                        }
                        self.set_status(401)
                        return dict_response
                    self.phanterpwa_current_client = q
                    return f(self, *args, **kargs)
                else:
                    if q:
                        q.delete_record()
                        self.DALDatabase.commit()
                    msg = "The phanterpwa-client-token is invalid!"
                    dict_response = {
                        'status': 'Unauthorized',
                        'code': 401,
                        'specification': 'client deleted',
                        'message': msg,
                        'i18n': {
                            'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                        }
                    }
                    self.set_status(401)
                    return dict_response
            else:
                return f(self, *args, **kargs)
        return check_client_token_decorator
    return decorator


def check_cas_token(ignore_locked=True):
    def decorator(f):
        @wraps(f)
        def check_cas_token_decorator(self, *args, **kargs):
            self.phanterpwa_cas_token_checked = None
            self.phanterpwa_cas_authorization = self.request.headers.get('phanterpwa-cas-authorization')
            if not self.phanterpwa_cas_authorization:
                msg = 'CAS authorization is not in the header. "phanterpwa-cas-authorization"'
                dict_response = {
                    'status': 'Bad Request',
                    'code': 400,
                    'message': msg,
                    'i18n': {
                        'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                    }
                }
                self.set_status(400)
                return dict_response
            t = Serialize(
                self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                self.projectConfig['BACKEND'][self.app_name]['default_time_cas_token_expire']
            )
            q = self.DALDatabase(self.DALDatabase.apps_authorization.authorization == self.phanterpwa_cas_authorization).select().first()
            if q:
                token_content = None
                try:
                    token_content = t.loads(self.phanterpwa_cas_authorization)
                except BadSignature:
                    token_content = None
                except SignatureExpired:
                    token_content = None
                if token_content:
                    if "user_agent" in token_content and "id_client" in token_content:
                        if token_content['app_name'] == q.app_name:
                            self.phanterpwa_cas_token_checked = token_content

            if self.phanterpwa_cas_token_checked:
                return f(self, *args, **kargs)
            else:
                if q:
                    q.delete_record()
                    self.DALDatabase.commit()
                msg = "The phanterpwa-cas-authorization is invalid!"
                dict_response = {
                    'status': 'Unauthorized',
                    'code': 401,
                    'specification': 'cas deleted',
                    'message': msg,
                    'i18n': {
                        'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                    }
                }
                self.set_status(401)
                return dict_response

        return check_cas_token_decorator
    return decorator


def check_url_token(ignore_user_agent=False):
    def decorator(f):
        @wraps(f)
        def check_url_token_decorator(self, *args, **kargs):

            dict_arguments = {k: self.request.arguments.get(k)[0].decode('utf-8') for k in self.request.arguments}
            self.phanterpwa_url_token_checked = None
            self.phanterpwa_url_token = dict_arguments.get('sign')
            self.phanterpwa_user_agent = self.request.headers.get('User-Agent')
            if not self.phanterpwa_url_token:
                msg = 'The URL token is not valid.'
                dict_response = {
                    'status': 'Bad Request',
                    'code': 400,
                    'message': msg,
                    'i18n': {
                        'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                    }
                }
                self.set_status(400)
                return dict_response
            t = URLSafeSerializer(
                self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                salt="url_secret_key"
            )
            token_content = None
            try:
                token_content = t.loads(self.phanterpwa_url_token)
            except BadSignature:
                token_content = None
            if token_content:
                if "user_agent" in token_content:
                    if ignore_user_agent:
                        self.phanterpwa_url_token_checked = token_content
                        q_user = self.DALDatabase(self.DALDatabase.auth_user.id == token_content.get("id_user")).select().first()
                        self.phanterpwa_current_user = q_user
                    else:
                        if token_content['user_agent'] == self.phanterpwa_user_agent:
                            q_user = self.DALDatabase(self.DALDatabase.auth_user.id == token_content.get("id_user")).select().first()
                            self.phanterpwa_current_user = q_user
                            self.phanterpwa_url_token_checked = token_content
            if self.phanterpwa_url_token_checked:
                return f(self, *args, **kargs)
            else:
                msg = "The URL token is invalid!"
                dict_response = {
                    'status': 'Forbidden',
                    'code': 403,
                    'message': msg,
                    'i18n': {
                        'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                    }
                }
                self.set_status(403)
                return dict_response
        return check_url_token_decorator
    return decorator


def check_public_csrf_token(form_identify=None, ignore_locked=True):
    def decorator(f):
        @wraps(f)
        @check_client_token(ignore_locked=ignore_locked)
        def check_csrf_token_decorator(self, *args, **kargs):
            dict_arguments = {k: self.request.arguments.get(k)[0].decode('utf-8') for k in self.request.arguments}
            self.phanterpwa_csrf_token_content = None
            self.phanterpwa_csrf_token_record = None
            self.phanterpwa_csrf_token = dict_arguments.get("csrf_token")
            self.phanterpwa_user_agent = self.request.headers.get('User-Agent')
            self.phanterpwa_remote_ip = self.request.headers.get("X-Real-IP") or \
                self.request.headers.get("X-Forwarded-For") or \
                self.request.remote_ip
            if not self.phanterpwa_csrf_token:
                msg = 'The CSRF token is not in form. "csrf_token"'
                dict_response = {
                    'status': 'Bad Request',
                    'code': 400,
                    'message': msg,
                    'i18n': {
                        'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                    }
                }
                self.set_status(400)
                return dict_response
            t = Serialize(
                "csrf-{0}".format(self.projectConfig['BACKEND'][self.app_name]['secret_key']),
                self.projectConfig['BACKEND'][self.app_name]['default_time_csrf_token_expire']
            )
            token_content = None
            try:
                token_content = t.loads(self.phanterpwa_csrf_token)
            except BadSignature:
                token_content = None
            except SignatureExpired:
                token_content = None
            if token_content:
                if 'id' in token_content:
                    q = self.DALDatabase(self.DALDatabase.csrf.id == token_content["id"]).select().first()
                    if q:
                        self.phanterpwa_csrf_token_record = q
                        if (q.token == self.phanterpwa_csrf_token) and\
                                self.phanterpwa_user_agent == q.user_agent and\
                                self.phanterpwa_remote_ip == q.ip:
                            self.phanterpwa_csrf_token_content = token_content
                            if form_identify:
                                def on_finish(handler_instance):
                                    data = handler_instance._final_response_on_success
                                    handler_instance.phanterpwa_csrf_token_record.update_record(
                                        success_response=json.dumps(data)
                                    )
                                    handler_instance.DALDatabase.commit()

                                self.phanterpwa_form_identify = token_content["form_identify"]
                                if isinstance(form_identify, str) and form_identify == self.phanterpwa_form_identify:
                                    if q.status_first_check == "OK" and q.success_response is not None:
                                        msg = "The form has already been processed"
                                        trans_msg = self.i18nTranslator.T(msg)
                                        dict_response = json.loads(q.success_response)
                                        dict_response['message'] = msg
                                        s_code = dict_response.get("code", 202)
                                        self.set_status(s_code)
                                        return dict_response
                                    self.on_finish = on_finish
                                    q.update_record(
                                        status_first_check="OK"
                                    )
                                    self.DALDatabase.commit()
                                    return f(self, *args, **kargs)
                                elif isinstance(form_identify, (list, tuple)) and \
                                        self.phanterpwa_form_identify in form_identify:
                                    if q.status_first_check == "OK" and q.success_response is not None:
                                        msg = "The form has already been processed"
                                        trans_msg = self.i18nTranslator.T(msg)
                                        dict_response = json.loads(q.success_response)
                                        dict_response['message'] = msg
                                        s_code = dict_response.get("code", 202)
                                        self.set_status(s_code)
                                        return dict_response
                                    self.on_finish = on_finish
                                    q.update_record(
                                        status_first_check="OK"
                                    )
                                    self.DALDatabase.commit()
                                    return f(self, *args, **kargs)
                                else:
                                    msg = "".join([
                                        "The crsf token is invalid! ",
                                        "The csrf token created for \"{0}\"",
                                        " is being used for a request that only accepts \"{1}\"."
                                    ])

                                    trans_msg = self.i18nTranslator.T(msg)
                                    msg = msg.format(str(self.phanterpwa_form_identify), str(form_identify))
                                    trans_msg = trans_msg.format(str(self.phanterpwa_form_identify), str(form_identify))
                                    dict_response = {
                                        'status': 'Bad Request',
                                        'code': 400,
                                        'message': msg,
                                        'i18n': {
                                            'message': trans_msg
                                        }
                                    }
                                    self.set_status(400)
                                    q.update_record(
                                        status_first_check="INVALID!"
                                    )
                                    return dict_response
                            q.update_record(
                                status_first_check="OK"
                            )
                            self.DALDatabase.commit()
                            return f(self, *args, **kargs)
                        else:
                            q.update_record(
                                status_first_check="INVALID!"
                            )
                            self.DALDatabase.commit()
                            msg = "The crsf token is invalid! The client has an unstable address."
                            dict_response = {
                                'status': 'Bad Request',
                                'code': 400,
                                'message': msg,
                                'i18n': {
                                    'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                                }
                            }
                            self.set_status(400)
                            return dict_response
            msg = "The crsf token is invalid!"
            dict_response = {
                'status': 'Bad Request',
                'code': 400,
                'message': msg,
                'i18n': {
                    'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                }
            }

            self.set_status(400)
            return dict_response
        return check_csrf_token_decorator
    return decorator


def check_user_token(ignore_activation=False):
    def decorator(f):
        @wraps(f)
        @check_client_token(ignore_locked=False)
        def check_user_token_decorator(self, *args, **kargs):
            def phanterpwa_current_user_has_role(obj, role):
                if not obj.phanterpwa_current_user_groups:
                    return False
                if isinstance(role, (list, tuple, set)):
                    if isinstance(role, (list, tuple)):
                        role = set(role)
                    auth_user_roles = set([x.role for x in obj.phanterpwa_current_user_groups])
                    if role.intersection(auth_user_roles):
                        return True
                    else:
                        return False
                elif isinstance(role, str):
                    return phanterpwa_current_user_has_role(obj, set([role]))
            self.phanterpwa_current_user_has_role = lambda role: phanterpwa_current_user_has_role(self, role)

            def add_role(obj, role):
                db = obj.DALDatabase()
                if not obj.phanterpwa_current_user_has_role(role) and obj.phanterpwa_current_user:
                    s_role = db(db.auth_group.role == role).select().first()
                    if s_role:
                        id_role = s_role.id
                    else:
                        id_role = db.auth_group.insert(
                            role=role,
                            grade=1
                        )
                    db.auth_membership.insert(
                        auth_user=obj.phanterpwa_current_user.id,
                        auth_group=id_role
                    )
                    db.commit()
                q_user_groups = db(
                    (db.auth_membership.auth_user == obj.phanterpwa_current_user.id) &
                    (db.auth_membership.auth_group == db.auth_group.id)
                ).select(
                    db.auth_group.id,
                    db.auth_group.role,
                    orderby=~db.auth_group.grade
                )
                if q_user_groups:
                    obj.phanterpwa_current_user_groups = q_user_groups
                else:
                    obj.phanterpwa_current_user_groups = None
                return obj.phanterpwa_current_user_groups
            self.add_role = lambda role: add_role(self, role)

            if not hasattr(self, "phanterpwa_user_token_checked"):
                self.phanterpwa_user_token_checked = None
                self.phanterpwa_current_user = None
                self.phanterpwa_current_user_groups = None
                self.phanterpwa_client_token = self.request.headers.get('phanterpwa-client-token')
                self.phanterpwa_authorization = self.request.headers.get('phanterpwa-authorization')
                id_user = None
                if self.phanterpwa_client_token and self.phanterpwa_authorization:
                    t = Serialize(
                        self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                        self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
                    )
                    token_content = None
                    try:
                        token_content = t.loads(self.phanterpwa_authorization)
                    except BadSignature:
                        token_content = None
                    except SignatureExpired:
                        token_content = None
                    if token_content and 'id' in token_content:
                        id_user = token_content['id']
                        self.phanterpwa_user_token_checked = token_content
                if id_user:
                    q_user = self.DALDatabase(self.DALDatabase.auth_user.id == id_user).select().first()
                    self.phanterpwa_current_user = q_user
                    q_user_groups = self.DALDatabase(
                        (self.DALDatabase.auth_membership.auth_user == self.phanterpwa_current_user.id) &
                        (self.DALDatabase.auth_membership.auth_group == self.DALDatabase.auth_group.id)
                    ).select(
                        self.DALDatabase.auth_group.id,
                        self.DALDatabase.auth_group.role,
                        orderby=~self.DALDatabase.auth_group.grade
                    )
                    if q_user_groups:
                        self.phanterpwa_current_user_groups = q_user_groups
                    q_client = self.DALDatabase(
                        (self.DALDatabase.client.auth_user == id_user) &
                        (self.DALDatabase.client.token == self.phanterpwa_client_token)
                    ).select().first()
                    if q_user and q_client:
                        if not q_user.activated and not ignore_activation:
                            msg = "The user token is invalid!"
                            dict_response = {
                                'status': 'Forbidden',
                                'code': 403,
                                'message': msg,
                                'i18n': {
                                    'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                                }
                            }
                            self.set_status(403)
                            return dict_response
                        else:

                            if not q_user.permit_mult_login:
                                r_client = self.DALDatabase(
                                    (self.DALDatabase.client.auth_user == id_user) &
                                    (self.DALDatabase.client.token != self.phanterpwa_client_token)
                                ).select()
                                if r_client:
                                    r_client = self.DALDatabase(
                                        (self.DALDatabase.client.auth_user == id_user) &
                                        (self.DALDatabase.client.token != self.phanterpwa_client_token)
                                    ).delete()
                            self.DALDatabase.commit()
                            return f(self, *args, **kargs)
                    else:
                        if q_client:
                            q_client.delete_record()
                        self.DALDatabase.commit()
                        msg = "The user token is invalid!"
                        dict_response = {
                            'status': 'Forbidden',
                            'code': 403,
                            'message': msg,
                            'i18n': {
                                'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                            }
                        }
                        self.set_status(403)
                        return dict_response
                else:
                    msg = "The user token is invalid!"
                    dict_response = {
                        'status': 'Forbidden',
                        'code': 403,
                        'message': msg,
                        'i18n': {
                            'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                        }
                    }
                    self.set_status(403)
                    return dict_response
            else:
                return f(self, *args, **kargs)
        return check_user_token_decorator
    return decorator


def check_private_csrf_token(form_identify=None, ignore_activation=False):
    def decorator(f):
        @wraps(f)
        @check_user_token(ignore_activation)
        @check_public_csrf_token(form_identify)
        def check_csrf_token_decorator(self, *args, **kargs):
            current_user = self.phanterpwa_csrf_token_content.get("user", None)
            if current_user and current_user == self.phanterpwa_current_user.id:
                return f(self, *args, **kargs)

            msg = "The crsf token is invalid!"
            dict_response = {
                'status': 'Bad Request',
                'code': 400,
                'message': msg,
                'i18n': {
                    'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                }
            }
            self.set_status(400)
            return dict_response
        return check_csrf_token_decorator
    return decorator


def requires_authentication(users_id=None, users_email=None, roles_id=None, roles_name=None):
    def decorator(f):
        @wraps(f)
        @check_user_token()
        def requires_authentication_decorator(self, *args, **kargs):
            if any([users_id, users_email, roles_id, roles_name]):
                msg = "User does not have sufficient privileges!"
                if isinstance(users_id, (list, tuple)):
                    if self.phanterpwa_current_user.id in users_id:
                        return f(self, *args, **kargs)
                if isinstance(users_id, int):
                    if self.phanterpwa_current_user.id == users_id:
                        return f(self, *args, **kargs)
                if isinstance(users_email, (list, tuple)):
                    if self.phanterpwa_current_user.email in users_email:
                        return f(self, *args, **kargs)
                if isinstance(users_email, str):
                    if users_email == self.phanterpwa_current_user.email:
                        return f(self, *args, **kargs)
                if self.phanterpwa_current_user_groups:
                    if isinstance(roles_id, int):
                        for x in self.phanterpwa_current_user_groups:
                            if x.id == roles_id:
                                return f(self, *args, **kargs)
                    if isinstance(roles_name, str):
                        for x in self.phanterpwa_current_user_groups:
                            if x.role == roles_name:
                                return f(self, *args, **kargs)
                    if isinstance(roles_id, (list, tuple)):
                        for x in self.phanterpwa_current_user_groups:
                            if x.id in roles_id:
                                return f(self, *args, **kargs)
                    if isinstance(roles_name, (list, tuple)):
                        for x in self.phanterpwa_current_user_groups:
                            if x.role in roles_name:
                                return f(self, *args, **kargs)

                msg = "User does not have sufficient privileges!"

                dict_response = {
                    'status': 'Unauthorized',
                    'code': 403,
                    'message': msg,
                    'i18n': {
                        'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                    }
                }
                self.set_status(403)
                return dict_response
            else:
                return f(self, *args, **kargs)

        return requires_authentication_decorator
    return decorator


def requires_no_authentication(ids=None, ignore_locked=True):
    def decorator(f):
        @wraps(f)
        @check_client_token(ignore_locked=ignore_locked)
        def requires_no_authentication_decorator(self, *args, **kargs):
            self.phanterpwa_current_user = None
            self.phanterpwa_client_token = self.request.headers.get('phanterpwa-client-token')
            self.phanterpwa_authorization = self.request.headers.get('phanterpwa-authorization')
            id_user = None
            if self.phanterpwa_client_token and self.phanterpwa_authorization:
                t = Serialize(
                    self.projectConfig['BACKEND'][self.app_name]['secret_key'],
                    self.projectConfig['BACKEND'][self.app_name]['default_time_user_token_expire']
                )
                token_content = None
                try:
                    token_content = t.loads(self.phanterpwa_authorization)
                except BadSignature:
                    token_content = None
                except SignatureExpired:
                    token_content = None
                if token_content and 'id' in token_content:
                    id_user = token_content['id']
                    self.phanterpwa_user_token_checked = token_content
            if id_user:
                q_user = self.DALDatabase(self.DALDatabase.auth_user.id == id_user).select().first()
                self.phanterpwa_current_user = q_user
                q_client = self.DALDatabase(
                    (self.DALDatabase.client.auth_user == id_user) &
                    (self.DALDatabase.client.token == self.phanterpwa_client_token)
                ).select().first()
                if q_user and q_client:
                    if not q_user.permit_mult_login:
                        r_client = self.DALDatabase(
                            (self.DALDatabase.client.auth_user == id_user) &
                            (self.DALDatabase.client.token != self.phanterpwa_client_token)
                        ).select()
                        if r_client:
                            r_client = self.DALDatabase(
                                (self.DALDatabase.client.auth_user == id_user) &
                                (self.DALDatabase.client.token != self.phanterpwa_client_token)
                            ).delete()
                    self.DALDatabase.commit()
                    msg = "The user must logout!"
                    dict_response = {
                        'status': 'Forbidden',
                        'code': 403,
                        'message': msg,
                        'i18n': {
                            'message': self.i18nTranslator.T(msg) if self.i18nTranslator else msg
                        }
                    }
                    self.set_status(403)
                    return dict_response
                else:
                    return f(self, *args, **kargs)
            else:
                return f(self, *args, **kargs)
        return requires_no_authentication_decorator
    return decorator



import os
import glob

current_path = os.path.dirname(__file__)
__all__ = [os.path.basename(x) for x in glob.glob(os.path.join(current_path, "*"))
    if os.path.isdir(x) and not os.path.basename(x).startswith("_")]
if __name__ == '__main__':
    print(__all__)



import sass
import os
from phanterpwa.helpers import (
    DIV,
    HTML,
    HEAD,
    BODY,
    META,
    TITLE,
    SPAN,
    XML,
    STYLE,
    TABLE,
    TBODY,
    TR,
    TD,
    P,
    BR,
    CONCATENATE
)


http_equiv = {
    "_http-equiv": "Content-Type",
    "_content": "text/html; charset=UTF-8"
}

css = ""
with open(os.path.join(os.path.dirname(__file__), "email.sass"), "r") as f:
    c = f.read()
    css = sass.compile(string=c, indented=True, output_style="compressed")


html = HTML(
    HEAD(
        TITLE("{{app_name}}"),
        META(**http_equiv),
        META(
            _name="viewport",
            _content="width=device-width"
        ),
        STYLE(css)
    ),
    BODY(
        SPAN(
            "Use this temporary password to create a new password:",
            " {{password}}.",
            _class="preheader"
        ),
        TABLE(
            TR(
                TD(
                    XML("&nbsp;")
                ),
                TD(
                    DIV(
                        TABLE(
                            TR(
                                TD(
                                    TABLE(
                                        TR(
                                            TD(
                                                P("Greetings {{user_name}},"),
                                                P("A new password was requested on {{app_name}}, a temporary password was sent (Expires in {{time_expires}}), use it to add a new password."),
                                                TABLE(
                                                    TR(
                                                        TD(
                                                            TABLE(
                                                                TBODY(
                                                                    TR(
                                                                        TD(
                                                                            DIV("{{password}}", _class="code")
                                                                        ),
                                                                    ),
                                                                ),
                                                                _role="presentation",
                                                                _border="0",
                                                                _cellpadding="0",
                                                                _cellspacing="0",
                                                            ),
                                                            _align="center"
                                                        ),
                                                    ),
                                                    _role="presentation",
                                                    _border="0",
                                                    _cellpadding="0",
                                                    _cellspacing="0",
                                                    _class="btn btn-primary"
                                                ),
                                                P("If you have not made this request, you do not have to do anything."),
                                                P("Any problem please contact support."),
                                            ),
                                        ),
                                        _role="presentation",
                                        _border="0",
                                        _cellpadding="0",
                                        _cellspacing="0"
                                    ),
                                    _class="wrapper"
                                ),
                            ),
                            _role="presentation",
                            _class="main"
                        ),
                        DIV(
                            TABLE(
                                TR(
                                    TD(
                                        SPAN(
                                            "{{copyright}}",
                                            _class="apple-link"
                                        ),
                                        BR(),
                                        XML("Do you like this app? <a href=\"{{link_to_your_page}}\">visit our page!</a>."),
                                        _class="content-block",
                                    )
                                ),
                                _role="presentation",
                                _border="0",
                                _cellpadding="0",
                                _cellspacing="0"
                            ),
                            _class="footer"
                        ),
                        _class="content"
                    ),
                    _class="container"
                ),
                TD(XML("&nbsp;")),
            ),
            _role="presentation",
            _border="0",
            _cellpadding="0",
            _cellspacing="0",
            _class="body"
        )
    )
)
text = CONCATENATE(
    "Greetings {{user_name}},",
    "\n\n",
    "A new password was requested on {{app_name}}, a temporary password was sent (Expires in {{time_expires}}), use it to add a new password.",
    "\n\n",
    "{{password}}",
    "\n\n",
    "If you have not made this request, you do not have to do anything.",
    "\n",
    "Any problem please contact support.",
    "\n\n",
    "{{copyright}}",
    "\n",
    "{{link_to_your_page}}"
)


import phanterpwa.frontend.gatehandler as gatehandler
import phanterpwa.frontend.helpers as helpers
import phanterpwa.frontend.forms as forms
import phanterpwa.frontend.preloaders as preloaders
from org.transcrypt.stubs.browser import __pragma__
__pragma__('alias', "jQuery", "$")
__pragma__('skip')
jQuery = window = 0
__pragma__('noskip')

DIV = helpers.XmlConstructor.tagger("div")
FORM = helpers.XmlConstructor.tagger("form")
SPAN = helpers.XmlConstructor.tagger("span")
IMG = helpers.XmlConstructor.tagger("img", True)
I = helpers.XmlConstructor.tagger("I")
I18N = helpers.I18N
CONCATENATE = helpers.CONCATENATE
H3 = helpers.XmlConstructor.tagger("h3")

html_base = CONCATENATE(
    DIV(
        DIV(
            DIV(
                DIV(I18N("Ops!!!", **{"_pt-br": "Ops!!!"}), _class="phanterpwa-breadcrumb"),
                _class="phanterpwa-breadcrumb-wrapper"
            ),
            _class="phanterpwa-container container"),
        _class='title_page_container card'
    ),
    DIV(
        DIV(
            DIV(
                DIV(
                    DIV(
                        IMG(
                            _class='image-warnings'
                        ),
                        _class="image-warnings-container"
                    ),
                    DIV(_id='content-warning'),
                    _class='content-warnings'
                ),
                _class='warnings-container phanterpwa-card-container card'
            ),
            _class="new-container"
        ),
        _class="phanterpwa-container container"
    )
)


class Error_404(gatehandler.ErrorHandler):
    def start(self):
        html = html_base.jquery()
        html.find(".image-warnings").attr(
            "src", "/static/{0}/images/warning.png".format(window.PhanterPWA.VERSION))
        html.find("#content-warning").html("ERROR 404")
        jQuery("#main-container").html(html)


class Error_401(gatehandler.ErrorHandler):
    def start(self):
        html = CONCATENATE(
            DIV(
                DIV(
                    DIV(
                        DIV(I18N("Authentication required", **{"_pt-br": "Necessário Autenticar-se"}), _class="phanterpwa-breadcrumb"),
                        _class="phanterpwa-breadcrumb-wrapper"
                    ),
                    _class="phanterpwa-container container"),
                _class='title_page_container card'
            ),
            DIV(
                DIV(
                    DIV(
                        DIV(
                            DIV(
                                IMG(
                                    _class='image-warnings',
                                    _src="/static/{0}/images/warning.png".format(window.PhanterPWA.VERSION)
                                ),
                            ),
                            _class="image-warnings-container"
                        ),
                        DIV(
                            I18N(
                                "You need authentication to access this feature.",
                                **{
                                    "_pt-br": "Você precisa autenticar-se para tentar acessar este recurso."
                                }
                            ),
                            _id='content-warning',
                            _class='content-warnings'
                        ),
                        DIV(
                            DIV(
                                I18N("Login"),
                                _id="alternative_login_button",
                                _class="btn wave_on_click link"
                            ),
                            _class="button-container"
                        ),
                        _class='warnings-container card phanterpwa-card-container'
                    ),
                    _class="new-container"
                ),
                _class="phanterpwa-container container"
            )
        )
        html.html_to("#main-container")
        jQuery("#alternative_login_button").off("click.alternative_login_button").on(
            "click.alternative_login_button",
            lambda: window.PhanterPWA.Components["auth_user"].modal_login()
        )


class Error_403(gatehandler.ErrorHandler):
    def start(self):
        html = html_base.jquery()
        html.find(".image-warnings").attr(
            "src", "/static/{0}/images/warning.png".format(window.PhanterPWA.VERSION))
        html.find("#content-warning").html("ERROR 403")
        jQuery("#main-container").html(html)



# -*- coding: utf-8 -*-

from ..helpers import (
    DIV,
    LABEL,
    INPUT,
    I,
    SPAN,
    TEXTAREA,
    SELECT,
    OPTION,
    CONCATENATE,
    UL,
    LI,
    A,
    P
)
from ..xmlconstructor import XmlConstructor


class MaterializeInputText(XmlConstructor):

    def __init__(self, inputname, label, id_input=None, default=None, error=None, **attributes):
        if not default:
            default = ""
        else:
            default = str(default)
        if not error:
            error = ""
        else:
            error = str(error)
        if not id_input:
            id_input = "input-%s" % inputname
        attributes["_phanterpwa-materialize-input_name"] = inputname
        attributes["_phanterpwa-materialize-input_id"] = id_input
        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-input-text"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-input-text"
        self.input = INPUT(
            _name=inputname,
            _id=id_input,
            _class="form-control",
            _value=default,
            _type="text"
        )
        new_content = [
            DIV(I("check", _class="material-icons"),
                _id="phanterpwa-materialize-input-check-%s" % inputname,
                _class="phanterpwa-materialize-input-check"),
            DIV(
                self.input,
                LABEL(
                    label,
                    _for=id_input),
                _class="input-field"),
            DIV(error,
                _title=error,
                _id="phanterpwa-materialize-input-error-%s" % inputname,
                _class="phanterpwa-materialize-input-error%s" % (" actived" if error != "" else " deactived")),
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)

    def disable(self):
        self.input['_disabled'] = ""


class MaterializeChips(XmlConstructor):
    def __init__(self, inputname, label, id_input=None, default="[]", error=None, **attributes):
        if not default:
            default = ""
        else:
            default = str(default)
        if not error:
            error = ""
        else:
            error = str(error)
        if not id_input:
            id_input = "input-%s" % inputname
        attributes["_phanterpwa-materialize-input_name"] = inputname
        attributes["_phanterpwa-materialize-input_id"] = id_input
        if "_id" not in attributes:
            attributes['_id'] = "phanterpwa-materialize-chips-jquery_plugin-%s" % inputname
        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-input-chips"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-input-chips"

        new_content = [
            DIV(I("check", _class="material-icons"),
                _id="phanterpwa-materialize-input-check-%s" % inputname,
                _class="phanterpwa-materialize-input-check"),
            DIV(
                DIV(
                    _id="phanterpwa-materialize-input-chips_%s" % inputname,
                    _class="phanterpwa-materialize-input-chips"),
                INPUT(
                    _name=inputname,
                    _id=id_input,
                    _class="form-control",
                    _value=default,
                    _type="hidden"),
                LABEL(
                    label,
                    _for=id_input),
                _class="input-field"),
            DIV(error,
                _title=error,
                _id="phanterpwa-materialize-input-error-%s" % inputname,
                _class="phanterpwa-materialize-input-error%s" % (" actived" if error != "" else " deactived")),
            DIV(_id="phanterpwa-materialize-input-chips-options-%s" % inputname, _class="phanterpwa-materialize-input-chips-options")
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)


class MaterializeSelectWithHideInput(XmlConstructor):
    def __init__(self, inputname, label, id_input=None, default=None, error=None, options=None, **attributes):
        if not default:
            default = ""
        else:
            default = str(default)
        if not error:
            error = ""
        else:
            error = str(error)
        if not id_input:
            id_input = "input-%s" % inputname
        attributes["_phanterpwa-materialize-input_name"] = inputname
        self.general_id = id_input
        self.general_name = inputname
        self.id_button_add_new = "button-%s" % id_input
        self.button_add_new = None
        attributes["_phanterpwa-materialize-input_id"] = id_input
        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-input-text"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-input-text"
        if options is None:
            options = []
        elif isinstance(options, (list, tuple)):
            for x in options:
                if not isinstance(x, (OPTION)):
                    raise TypeError("elements of options must is OPTION instance")
        elif isinstance(options, (dict)):
            new_options = [OPTION("Escolha uma opção", _value=" ", _disabled="", _selected="")]
            for x in options:
                new_options.append(OPTION(options[x], _value="%s" % x))
            options = new_options
        else:
            raise TypeError("options must is tuple, list or None")

        self._html_select = SELECT(
            *options,
            _id="select-%s" % id_input,
            _class="phanterpwaformselect-withhiddeninput",
            _target_input=id_input)
        self._html_select_label = LABEL(label, _for="select-%s" % id_input)
        self._html_select_concatenate = CONCATENATE(
            self._html_select,
            self._html_select_label
        )
        self._html_check = DIV(
            I("check", _class="material-icons"),
            _id="phanterpwa-materialize-input-check-%s" % inputname,
            _class="phanterpwa-materialize-input-check actived-select")
        attr_btn_switch = {}
        attr_btn_switch["_target-switch"] = "switch-input-select-container-%s" % self.general_id
        attr_btn_switch["_target-check"] = "phanterpwa-materialize-input-check-%s" % self.general_name
        new_content = [
            self._html_check,
            DIV(
                DIV(
                    INPUT(
                        _name=inputname,
                        _id=id_input,
                        _class="form-control",
                        _value=default,
                        _type="text"),
                    LABEL(
                        label,
                        _for=id_input),
                    DIV(
                        I("details", _class="material-icons"),
                        _class="phanterpwa-materilize-button-show-hidde-input-new doinput waves-effect waves-teal btn link",
                        **attr_btn_switch),
                    _id="switch-input-%s" % id_input,
                    _class="input-field switch-input"),
                DIV(
                    self._html_select_concatenate,
                    _id="switch-select-%s" % id_input,
                    _class="input-field switch-select"),
                _id="switch-input-select-container-%s" % id_input,
                _class="switch-input-select-container actived-select"),
            DIV(error,
                _title=error,
                _id="phanterpwa-materialize-input-error-%s" % inputname,
                _class="phanterpwa-materialize-input-error%s" % (" actived" if error != "" else " deactived")),
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)

    def addOptionInSelect(self, value, label_value, **attributes):
        if "_value" in attributes:
            del attributes['_value']
        option = OPTION(label_value, _value=value, **attributes)
        self._html_select.content.append(option)

    def setButtonNew(self, label=None, **attributes):
        if label is None:
            label = I("add", _class="material-icons")
        attributes['_id'] = self.id_button_add_new
        attributes["_target-switch"] = "switch-input-select-container-%s" % self.general_id
        attributes["_target-check"] = "phanterpwa-materialize-input-check-%s" % self.general_name
        class_button = "phanterpwa-materilize-button-show-hidde-input-new waves-effect waves-teal btn link"
        self._html_check.attributes["_class"] = " ".join(["hasbutton", self._html_check.attributes["_class"]])
        if "_class" in attributes:
            attributes["_class"] = " ".join([attributes["_class"], class_button])
        else:
            attributes["_class"] = class_button
        self._html_select_concatenate.content = [
            DIV(
                DIV(
                    self._html_select,
                    self._html_select_label,
                    _class="phanterpwa-materialize-selectwithinput"),
                DIV(
                    DIV(label, **attributes),
                    _class="phanterpwa-materialize-selectwithinput-button"),
                _class="phanterpwa-materialize-select-and-button-container"
            )
        ]


class MaterializeInputTextMultiline(XmlConstructor):
    def __init__(self, inputname, label, id_input=None, default=None, error=None, **attributes):
        if not default:
            default = ""
        else:
            default = str(default)
        if not error:
            error = ""
        else:
            error = str(error)
        if not id_input:
            id_input = "input-%s" % inputname
        attributes["_phanterpwa-materialize-input_name"] = inputname
        attributes["_phanterpwa-materialize-input_id"] = id_input
        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-input-text"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-input-text"

        new_content = [
            DIV(I("check", _class="material-icons"),
                _id="phanterpwa-materialize-input-check-%s" % inputname,
                _class="phanterpwa-materialize-input-check"),
            DIV(
                TEXTAREA(
                    _name=inputname,
                    _id=id_input,
                    _class="form-control materialize-textarea",
                    _value=default,
                    _type="text"),
                LABEL(
                    label,
                    _for=id_input),
                _class="input-field"),
            DIV(error,
                _title=error,
                _id="phanterpwa-materialize-input-error-%s" % inputname,
                _class="phanterpwa-materialize-input-error%s" % (" actived" if error != "" else " deactived")),
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)


class MaterializeInputHidden(XmlConstructor):
    def __init__(self, inputname, label, id_input=None, default=None, error=None, **attributes):
        if not default:
            default = ""
        else:
            default = str(default)
        if not error:
            error = ""
        else:
            error = str(error)
        if not id_input:
            id_input = "input-%s" % inputname
        attributes["_phanterpwa-materialize-input_name"] = inputname
        attributes["_phanterpwa-materialize-input_id"] = id_input
        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-input-hidden"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-input-hidden"

        new_content = [
            DIV(I("check", _class="material-icons"),
                _id="phanterpwa-materialize-input-check-%s" % inputname,
                _class="phanterpwa-materialize-input-check"),
            DIV(
                INPUT(
                    _name=inputname,
                    _id=id_input,
                    _class="form-control",
                    _value=default,
                    _type="hidden"),
                LABEL(
                    label,
                    _for=id_input),
                _class="input-field"),
            DIV(error,
                _title=error,
                _id="phanterpwa-materialize-input-error-%s" % inputname,
                _class="phanterpwa-materialize-input-error%s" % (" actived" if error != "" else " deactived")),
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)


class MaterializeInputPassword(XmlConstructor):
    def __init__(self, inputname, label, id_input=None, default=None, error="", **attributes):
        if not default:
            default = ""
        else:
            default = str(default)
        if not error:
            error = ""
        else:
            error = str(error)
        if not id_input:
            id_input = "input-%s" % inputname
        attributes["_phanterpwa-materialize-input_name"] = inputname
        attributes["_phanterpwa-materialize-input_id"] = id_input
        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-input-password"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-input-password"

        new_content = [
            DIV(I("check", _class="material-icons"),
                _id="phanterpwa-materialize-input-check-%s" % inputname,
                _class="phanterpwa-materialize-input-check"),
            DIV(
                INPUT(
                    _name=inputname,
                    _id=id_input,
                    _class="form-control",
                    _value=default,
                    _type="password"),
                LABEL(
                    label,
                    _for=id_input),
                _class="input-field"),
            DIV(error,
                _title=error,
                _id="phanterpwa-materialize-input-error-%s" % inputname,
                _class="phanterpwa-materialize-input-error%s" % (" actived" if error != "" else " deactived")),
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)


class MaterializeInputCheckBox(XmlConstructor):
    def __init__(self, inputname, label, id_input=None, disabled=False, checked=False, filledIn=False, **attributes):
        if not id_input:
            id_input = "input-%s" % inputname
        attributes["_phanterpwa-materialize-input_name"] = inputname
        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-input-checkbox"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-input-checkbox"
        new_content = [
            DIV(
                P(
                    LABEL(
                        INPUT(
                            _name=inputname,
                            _id=id_input,
                            _class="form-control%s" % (" filled-in" if filledIn else ""),
                            _checked="checked" if checked else None,
                            _disabled="disabled" if disabled else None,
                            _type="checkbox"),
                        SPAN(label),
                        _for=id_input
                    )
                ),
                _class="input-field-checkbox"),
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)


class MaterializeButtonForm(XmlConstructor):
    def __init__(self, _id, label, **attributes):
        self.label = label
        initial_class = "phanterpwa-materialize-button-form-container"
        attributes["_id"] = _id
        self.button_attributes = attributes
        if "_class" in self.button_attributes:
            self.button_attributes["_class"] = " ".join([
                self.button_attributes['_class'].strip(),
                "btn phanterpwa-materialize-button-form link"])
        else:
            self.button_attributes["_class"] = "btn phanterpwa-materialize-button-form link"
        if "_title" not in self.button_attributes:
            if isinstance(self.label, str):
                self.button_attributes["_title"] = self.label

        XmlConstructor.__init__(self, 'div', False, _class=initial_class)
        self._update_content()

    def _update_content(self):
        attributes = self.button_attributes
        self.content = [
            DIV(
                DIV(self.label, **attributes),
                _class="button-form")
        ]




class MaterializeSearchBar(XmlConstructor):
    def __init__(
            self,
            inputname,
            fields_to_select=None,
            label="Pesquisar",
            field_select_label="Campo de pesquisa",
            **attributes
        ):
        self.field_select = field_select_label
        self.inputname = inputname
        id_input = "phanterpwa-materialize-search_bar-input-%s" % inputname

        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-search_bar"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-search_bar"
        self.select_search_container = CONCATENATE()
        self.select_search = None
        self.input_search = DIV(
            DIV(
                INPUT(
                    _name=inputname,
                    _id=id_input,
                    _class="form-control",
                    _type="text"),
                LABEL(
                    label,
                    _for=id_input),
                _class="input-field"),
            DIV(
                I("search", _class="material-icons"),
                _source_search=id_input,
                _source_select="phanterpwa-materialize-select-search-%s" % inputname,
                _id="phanterpwa-materialize-search_bar-button-%s" % inputname,
                _class="waves-effect waves-teal btn link materialize-search_bar-button"),

            _id='materialize-search_bar-input-and-button-%s' % inputname,
            _class='materialize-search_bar-input-and-button'
        )
        new_content = [
            CONCATENATE(
                DIV(
                    self.input_search,
                    self.select_search_container,
                    _class="row")
            )
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)

    def addOptionInSelect(self, value, label_value, **attributes):
        if not self.select_search_container.content:
            self.input_search.attributes["_class"] = " ".join([
                "has_select col s12 m8 l8",
                self.input_search.attributes["_class"]])
            self.select_search = SELECT(
                _id="phanterpwa-materialize-select-search-%s" % self.inputname,
                _class="phanterpwa-materialize-select-search")
            select_container = DIV(
                self.select_search,
                LABEL(self.field_select, _for="phanterpwa-materialize-select-search-%s" % self.inputname),
                _class="input-field")
            self.select_search_container.content = [
                DIV(
                    select_container,
                    _class="phanterpwa-materialize-search_bar-select col s12 m4 l4")
            ]
        self.select_search.append(OPTION(label_value, _value=value, **attributes))

    def showSelect(self):
        if not self.select_search_container.content and self.select_search is None:
            self.input_search["_class"] = " ".join([
                "has_select col s12 m8 l8",
                self.input_search.attributes["_class"]])
            self.select_search = SELECT(
                _id="phanterpwa-materialize-select-search-%s" % self.inputname,
                _class="phanterpwa-materialize-select-search")
            select_container = DIV(
                self.select_search,
                LABEL(self.field_select, _for="phanterpwa-materialize-select-search-%s" % self.inputname),
                _class="input-field")
            self.select_search_container.content = [
                DIV(
                    select_container,
                    _class="phanterpwa-materialize-search_bar-select col s12 m4 l4")
            ]


class MaterializeFloatButton(XmlConstructor):
    def __init__(self, icon_name="build", **attributes):
        if "_class" in attributes:
            new_class = " ".join([
                attributes["_class"].strip(),
                "phanterpwa-materialize-floating-action-button"
            ])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-floating-action-button"
        self.buttons_floating = UL()
        self.principal_icon = CONCATENATE(
            A(
                I(icon_name, _class="large material-icons"),
                _class="btn-floating btn-large blue waves-effect waves-light")
        )
        new_content = [
            DIV(
                self.principal_icon,
                self.buttons_floating,
                _class="fixed-action-btn click-to-toggle")
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)

    def addMaterializeActionButton(self, material_icon_name, color="blue", **attributes):
        if '_class' in attributes:
            attributes['_class'] = " ".join(["btn-floating %s" % color, attributes['_class']])
        else:
            attributes['_class'] = "btn-floating %s" % color
        self.buttons_floating.append(LI(A(I(material_icon_name, _class="material-icons"), **attributes)))

    def changePrincipalMaterializeActionButton(self, material_icon_name, color="blue", **attributes):
        if '_class' in attributes:
            attributes['_class'] = " ".join(["btn-floating %s" % color, attributes['_class']])
        else:
            attributes['_class'] = "btn-floating %s" % color
        self.principal_icon.content = [
            A(I(material_icon_name, _class="material-icons"), **attributes)]

    def addActionButton(self, personal_icon, color="blue", **attributes):
        if '_class' in attributes:
            attributes['_class'] = " ".join(["btn-floating %s" % color, attributes['_class']])
        else:
            attributes['_class'] = "btn-floating %s" % color
        self.buttons_floating.append(LI(A(personal_icon, **attributes)))

    def changePrincipalActionButton(self, personal_icon, color="blue", **attributes):
        if '_class' in attributes:
            attributes['_class'] = " ".join(["btn-floating %s" % color, attributes['_class']])
        else:
            attributes['_class'] = "btn-floating %s" % color
        self.principal_icon.content = [
            A(personal_icon, **attributes)]


class MaterializeTabs(XmlConstructor):
    def __init__(self, **attributes):
        if "_class" in attributes:
            new_class = " ".join([attributes["_class"].strip(), "phanterpwa-materialize-tabs row"])
            attributes["_class"] = new_class
        else:
            attributes["_class"] = "phanterpwa-materialize-tabs row"
        self.tabs_menu = UL(_class="tabs")
        self.tabs = CONCATENATE()
        self.tabs_elements = []
        new_content = [
            CONCATENATE(
                DIV(
                    self.tabs_menu,
                    _class="col s12"),
                self.tabs
            )
        ]
        XmlConstructor.__init__(self, 'div', False, *new_content, **attributes)

    def addTab(self, id_tab, title, content, actived=False, disabled=False, **attributes):
        class_base = "phanterpwa-materialize-tab-menu tab"
        if disabled:
            class_base = "phanterpwa-materialize-tab-menu tab disabled"
        if "_class" in attributes:
            attributes["_class"] = " ".join([class_base, attributes["_class"]])
        else:
            attributes["_class"] = " ".join([class_base, attributes["_class"]])
        element_menu = LI(A(title, _href="#%s" % id_tab), **attributes)
        element_tab = DIV(content, _id=id_tab, _class="col s12")
        self.tabs_elements.append([element_menu, element_tab])
        self.tabs_menu.content = []
        self.tabs.content = []
        for x in self.tabs_elements:
            self.tabs_menu.append(x[0])
            self.tabs.append(x[1])



import os
import sys
import json
import datetime
import phanterpwa
import configparser
from glob import glob
from os.path import (
    normpath,
    join,
    isdir,
    isfile,
    dirname,
    basename
)
from phanterpwa.samples.project_config_sample import project_config_sample

ENV_PYTHON = normpath(sys.executable)
VERSION_PYTHON = "{0}.{1}.{2}".format(sys.version_info[0], sys.version_info[1], sys.version_info[2])
PATH_PHANTERPWA = dirname(phanterpwa.__file__)
VERSION_PHANTERPWA = phanterpwa.__version__


class ProjectConfig():
    """
    Validates and creates a configuration file in json format based on api.ini and app.ini contained in the project's
    applications.

    To start when instantiating, enter the path to the file.

    :param dir_or_config_file: Project dir or config.json file

    Example:
        >>> import os
        >>> import phanterpwa
        >>> from phanterpwa.configer import ProjectConfig
        >>> test_folder = os.path.join(os.path.dirname(phanterpwa.__file__), "tests", "test_configer_path", "project01")
        >>> cfg = ProjectConfig(test_folder)
    """

    def __init__(self, dir_or_config_file):
        self.project_config_sample = self._new_dict(project_config_sample)
        self._set_file(dir_or_config_file)

    def _new_dict(self, d):
        for x in d:
            if isinstance(d[x], dict):
                d[x] = self._new_dict(d[x])
        return dict(d)

    @property
    def file(self):
        """gets the path of the current configuration file

        Example:
            >>> import os
            >>> import phanterpwa
            >>> from phanterpwa.configer import ProjectConfig
            >>> test_folder = os.path.join(os.path.dirname(phanterpwa.__file__), "tests", "test_configer_path", "project01")
            >>> cfg = ProjectConfig(test_folder)
            >>> cfg.file
            ../phanterpwa/tests/test_configer_path/project01/config.json
        """
        return self._file

    @property
    def project_ini(self):
        """returns the configparser instance of the "project.ini" file
        """
        return self._ini_project

    @property
    def project_secret_ini(self):
        """returns the configparser instance of the "secret.ini" file
        """
        return self._ini_secret

    @property
    def backend_ini(self):
        """returns a dictionary with the configparser instances of the "app.ini" files for each backend application.
        The keys are the folder name for each application.
        """
        return self._ini_apps_backend

    @property
    def frontend_ini(self):
        """returns a dictionary with the configparser instances of the "app.ini" files for each frontend application.
        The keys are the folder name for each application.
        """
        return self._ini_apps_frontend

    @property
    def backend_secret_ini(self):
        """returns a dictionary with the configparser instances of the "app.ini" files for each backend application.
        The keys are the folder name for each application.
        """
        return self._ini_secret_apps_backend

    @property
    def frontend_secret_ini(self):
        """returns a dictionary with the configparser instances of the "app.ini" files for each frontend application.
        The keys are the folder name for each application.
        """
        return self._secret_ini_apps_frontend

    def _process(self):
        cfg = {}
        with open(self._file, 'r', encoding="utf-8") as f:
            cfg = json.load(f)
        if isinstance(cfg, dict):
            identify = cfg.get("CONFIG_INDENTIFY", None)

            if not identify:
                raise ValueError("The config file don't have the CONFIG_INDENTIFY key.")
            required_keys = self.project_config_sample.keys()
            for r in required_keys:
                if r not in cfg:
                    cfg[r] = {}
            cfg["CONFIG_INDENTIFY"] = "project_config"
            cfg["ENVIRONMENT"] = {
                "python": ENV_PYTHON,
                "path": normpath(dirname(ENV_PYTHON)),
                "python_version": VERSION_PYTHON,
                "phanterpwa_version": VERSION_PHANTERPWA
            }
            self._ini_secret = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})
            self._ini_secret.read(join(cfg["PROJECT"]["path"], 'secret.ini'), encoding='utf-8')
            sections_secret = self._ini_secret.sections()

            self._ini_project = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})
            self._ini_project.read(join(cfg["PROJECT"]["path"], 'project.ini'), encoding='utf-8')
            sections_project = self._ini_project.sections()
            required_on_ini = [
                "title",
                "version",
                "author",
                "debug",
                "minify",
                "packaged"
            ]
            if "PROJECT" in sections_project:
                cfg["PROJECT"]["path"] = normpath(dirname(join(self._file)))
                cfg["PROJECT"]["name"] = basename(cfg["PROJECT"]["path"])
                sections_project.pop(sections_project.index("PROJECT"))
                for k in required_on_ini:
                    if k == "debug" or k == "packaged" or k == "minify":
                        cfg["PROJECT"][k] = self._ini_project["PROJECT"].getboolean(
                            k, self.project_config_sample["PROJECT"][k])
                    elif k == "baseport":
                        v = self._ini_project["PROJECT"].get(
                            k, self.project_config_sample["PROJECT"][k])
                        if str(v).isdigit():
                            cfg["PROJECT"][k] = int(v)
                    else:
                        cfg["PROJECT"][k] = self._ini_project["PROJECT"].get(
                            k, self.project_config_sample["PROJECT"][k])
                for i in self._ini_project.items("PROJECT"):
                    cfg["PROJECT"][i[0]] = self._dict_value(i[1])
            else:
                cfg["PROJECT"] = self.project_config_sample["PROJECT"]
                cfg["PROJECT"]["path"] = normpath(dirname(join(self._file)))
                cfg["PROJECT"]["name"] = basename(cfg["PROJECT"]["path"])
                self._ini_project["PROJECT"] = {}
            for k in required_on_ini:
                self._ini_project["PROJECT"][k] = str(cfg["PROJECT"][k])
            if "compilation" not in cfg["PROJECT"]:
                cfg["PROJECT"]["compilation"] = 0
                cfg["PROJECT"]["date_compilation"] = str(datetime.datetime.now())
                cfg["PROJECT"]["versioning"] = "{0}.{1}".format(
                    cfg["PROJECT"].get("version", "0.0.1"),
                    0
                )
            if "baseport" not in cfg["PROJECT"]:
                cfg["PROJECT"]["baseport"] = self.project_config_sample["PROJECT"]["baseport"]
            if "basehost" not in cfg["PROJECT"]:
                cfg["PROJECT"]["basehost"] = self.project_config_sample["PROJECT"]["basehost"]

            current_port = cfg["PROJECT"]["baseport"]
            ports = set()

            self._ini_apps_backend = dict()
            self._ini_secret_apps_backend = dict()
            apps_backend_list = []
            for a in glob(join(cfg["PROJECT"]["path"], "backapps", "*")):
                if isdir(a):
                    app_name = basename(a)
                    apps_backend_list.append(app_name)
                    if isfile(join(a, "app.ini")):
                        ini_app = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})
                        ini_app.read(join(a, "app.ini"), encoding='utf-8')
                        app_sections = ini_app.sections()
                        required_on_ini = [
                            "title"
                        ]
                        if "APP" in app_sections:
                            cfg["BACKEND"][app_name] = {
                                "default_time_user_token_expire": 7200,
                                "default_time_user_token_expire_remember_me": 2592000,
                                "default_time_csrf_token_expire": 4200,
                                "default_time_temporary_password_expire": 4200,
                                "timeout_to_resend_temporary_password_mail": 3900,
                                "default_time_client_token_expire": 63072000,
                                "default_time_activation_code_expire": 3600,
                                "wait_time_to_try_activate_again": 3900,
                                "timeout_to_resend_activation_email": 300,
                                "timeout_to_resign": 600,
                                "timeout_to_next_login_attempt": 4200,
                                "max_login_attempts": 5,
                                "max_activation_attempts": 5,
                                "default_time_user_token_expire": 7200,
                                "default_time_two_factor_code_expire": 3600
                            }

                            for k in required_on_ini:
                                v = ini_app["APP"].get(k, None)
                                if v is None:
                                    if k == "port" or k == "host":
                                        pass
                                    else:
                                        cfg["BACKEND"][app_name][k] = self.project_config_sample["BACKEND"]["sample_app"][k]
                                        ini_app["APP"][k] = cfg["BACKEND"][app_name][k]
                                else:
                                    if v.isdigit():
                                        if k == "port" or k == "host":
                                            pass
                                        else:
                                            cfg["BACKEND"][app_name][k] = int(v)
                                    elif v == "false":
                                        cfg["BACKEND"][app_name][k] = False
                                    elif v == "true":
                                        cfg["BACKEND"][app_name][k] = True
                                    else:
                                        cfg["BACKEND"][app_name][k] = v
                            for i in ini_app.items("APP"):
                                cfg["BACKEND"][app_name][i[0]] = self._dict_value(i[1])
                        else:
                            ini_app["APP"] = {}
                            cfg["BACKEND"][app_name] = self.project_config_sample["BACKEND"]["sample_app"]
                        for k in required_on_ini:
                            ini_app["APP"][k] = str(cfg["BACKEND"][app_name][k])
                        cfg["BACKEND"][app_name]["host"] = cfg["PROJECT"]["basehost"] or "127.0.0.1"
                        while current_port in ports:
                            current_port += 1
                        else:
                            ports.add(current_port)
                            cfg["BACKEND"][app_name]["port"] = current_port
                        # with open(join(a, "app.ini"), 'w', encoding="utf-8") as configfile:
                        #     ini_app.write(configfile)

                        self._ini_apps_backend[app_name] = ini_app

                    if isfile(join(a, "secret.ini")):
                        ini_secret_app = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})
                        ini_secret_app.read(join(a, "secret.ini"), encoding='utf-8')
                        app_sections = ini_secret_app.sections()
                        if "APP" in app_sections:
                            if app_name in cfg["BACKEND"] and isinstance(cfg["BACKEND"][app_name], dict):
                                for i in ini_secret_app.items("APP"):
                                    cfg["BACKEND"][app_name][i[0]] = self._dict_value(i[1])
                        else:
                            ini_secret_app["APP"] = {}
                        # with open(join(a, "secret.ini"), 'w', encoding="utf-8") as configfile:
                        #     ini_secret_app.write(configfile)

                        self._ini_secret_apps_backend[app_name] = ini_secret_app
                    else:
                        ini_secret_app = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})
                        ini_secret_app["APP"] = {
                            'secret_key': os.urandom(12).hex()
                        }
                        with open(join(a, "secret.ini"), 'w', encoding="utf-8") as configfile:
                            ini_secret_app.write(configfile)

            current_apps = list(cfg['BACKEND'].keys())
            for ha in current_apps:
                if ha not in apps_backend_list:
                    del cfg['BACKEND'][ha]

            self._ini_apps_frontend = dict()
            self._ini_secret_apps_frontend = dict()
            apps_frontend_list = []

            for a in glob(join(cfg["PROJECT"]["path"], "frontapps", "*")):
                if isdir(a):
                    app_name = basename(a)
                    apps_frontend_list.append(app_name)
                    if isfile(join(a, "app.ini")):
                        ini_app = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})
                        ini_app.read(join(a, "app.ini"), encoding='utf-8')
                        app_sections = ini_app.sections()
                        required_on_ini = [
                            "title",
                            "description",
                            "transcrypt_main_file",
                            "styles_main_file",
                            "views_main_file"
                        ]
                        if "APP" in app_sections:
                            cfg["FRONTEND"][app_name] = {"build_folder":
                                normpath(join(cfg["PROJECT"]["path"], "frontapps", app_name, "www"))}

                            for k in required_on_ini:
                                v = ini_app["APP"].get(k, None)
                                if v is None:
                                    if k == "port" or k == "host":
                                        pass
                                    else:
                                        cfg["FRONTEND"][app_name][k] = self.project_config_sample["FRONTEND"]["sample_app"][k]
                                        ini_app["APP"][k] = cfg["FRONTEND"][app_name][k]
                                else:
                                    if v.isdigit():
                                        if k == "port" or k == "host":
                                            pass
                                        else:
                                            cfg["FRONTEND"][app_name][k] = int(v)
                                    elif v == "false":
                                        cfg["FRONTEND"][app_name][k] = False
                                    elif v == "true":
                                        cfg["FRONTEND"][app_name][k] = True
                                    else:
                                        cfg["FRONTEND"][app_name][k] = v
                            for i in ini_app.items("APP"):
                                cfg["FRONTEND"][app_name][i[0]] = self._dict_value(i[1])
                        else:
                            ini_app["APP"] = {}
                            cfg["FRONTEND"][app_name] = self.project_config_sample["FRONTEND"]["sample_app"]
                            cfg["FRONTEND"][app_name]["build_folder"] = \
                                normpath(join(cfg["PROJECT"]["path"], "frontapps", app_name, "www"))
                        for k in required_on_ini:
                            ini_app["APP"][k] = str(cfg["FRONTEND"][app_name][k])

                        cfg["FRONTEND"][app_name]["host"] = cfg["PROJECT"]["basehost"] or "127.0.0.1"
                        while current_port in ports:
                            current_port += 1
                        else:
                            ports.add(current_port)
                            cfg["FRONTEND"][app_name]["port"] = current_port
                        # with open(join(a, "app.ini"), 'w', encoding="utf-8") as configfile:
                        #     ini_app.write(configfile)

                        self._ini_apps_frontend[app_name] = ini_app

                    if isfile(join(a, "secret.ini")):
                        ini_secret_app = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})
                        ini_secret_app.read(join(a, "secret.ini"), encoding='utf-8')
                        app_sections = ini_secret_app.sections()
                        if "APP" in app_sections:
                            if app_name in cfg["FRONTEND"] and isinstance(cfg["FRONTEND"][app_name], dict):
                                for i in ini_secret_app.items("APP"):
                                    cfg["FRONTEND"][app_name][i[0]] = self._dict_value(i[1])
                        else:
                            ini_secret_app["APP"] = {}

                        # with open(join(a, "secret.ini"), 'w', encoding="utf-8") as configfile:
                        #     ini_secret_app.write(configfile)

                        self._ini_secret_apps_frontend[app_name] = ini_secret_app


            current_apps = list(cfg['FRONTEND'].keys())
            for ha in current_apps:
                if ha not in apps_frontend_list:
                    del cfg['FRONTEND'][ha]

            for p in sections_project:
                if p not in ["PROJECT"]:
                    cfg[p] = self._dict_value(dict(self._ini_project[p]))
            for p in sections_secret:
                if p in cfg:
                    secret_dict_values = self._dict_value(dict(self._ini_secret[p]))
                    for k in secret_dict_values:
                        cfg[p][k] = secret_dict_values[k]
                else:
                    cfg[p] = self._dict_value(dict(self._ini_secret[p]))

        else:
            raise TypeError(
                "The config file has incorrect content, expected dict type. Given: {0}".format(type(cfg)))

        self._config = self._new_dict(cfg)

    def _set_file(self, value):
        if isfile(value) and basename(value) == "config.json":
            self._file = value
            self._process()
            self.save()
        elif isdir(value):
            if not isfile(os.path.join(value, "config.json")):
                with open(os.path.join(value, "config.json"), "w", encoding="utf-8") as f:
                    f.write(json.dumps(self.project_config_sample, ensure_ascii=True, indent=2))
            self._file = os.path.join(value, "config.json")
            self._process()
            self.save()
        else:
            raise ValueError("The config file must be config.json file or a valid dir. Given: {0}".format(value))

    def _dict_value(self, d):
        if isinstance(d, dict):
            for x in d:
                d[x] = self._dict_value(d[x])
            return d
        elif isinstance(d, str):
            if d.isdigit():
                return int(d)
            elif d.lower() == "false":
                return False
            elif d.lower() == "true":
                return True
            elif d.lower() == "none":
                return None
            else:
                return d
        else:
            return d

    @property
    def config(self):
        """returns the current setting

        Example:
            >>> import json
            >>> import os
            >>> import phanterpwa
            >>> from phanterpwa.configer import ProjectConfig
            >>> test_folder = os.path.join(os.path.dirname(phanterpwa.__file__), "tests", "test_configer_path", "project01")
            >>> cfg = ProjectConfig(test_folder)
            >>> string_cfg = json.dumps(cfg.config, indent=2)
            >>> print(string_cfg)
            {
              "CONFIG_INDENTIFY": "project_config",
              "ENVIRONMENT": {
                "path": "",
                "python": ""
              },
              "PROJECT": {
                "name": "PhanterPWA",
                "title": "PhanterPWA",
                "version": "0.0.1",
                "author": "PhanterJR<phanterjr@conexaodidata.com.br>",
                "debug": true,
                "packaged": true
              },
              "BACKEND": {
              },
              "FRONTEND": {
              },
              "PATH": {
                "project": "{{PROJECT_FOLDER}}",
                "api": "{{PROJECT_FOLDER}}\\api",
                "app": "{{PROJECT_FOLDER}}\\app"
              },
              "EMAIL": {
                "server": "mail.yourservermail.com",
                "username": "username@yourservermail.com",
                "default_sender": "contato@conexaodidata.com.br",
                "password": "password",
                "port": 465,
                "use_tls": true,
                "use_ssl": true
              },
              "TRANSCRYPT": {
                "main_files": [
                  "{{PROJECT_FOLDER}}\\app\\scripts\\application\\application.py",
                  "{{PROJECT_FOLDER}}\\app\\scripts\\websocket\\websocket.py"
                ]
              },
              "API_SERVER": {
                "host": "127.0.0.1",
                "port": 8881
              },
              "APP_SERVER": {
                "host": "127.0.0.1",
                "port": 8882
              },
              "CONFIGJS": {
                "api_server_address": "http://127.0.0.1:8881",
                "api_websocket_address": "ws://127.0.0.1:8881/websocket",
                "timeout_to_resign": 600
              },
              "CONTENT_EMAILS": {
                "copyright": "Conex\u00e3o Didata \u00a9 2011-{{now}}",
                "link_to_your_site": "https://phanterpwa.conexaodidata.com.br"
              }
            }
        """
        return self._config

    def save(self):
        """saves changes made to configuration files
            Example:
            >>> import json
            >>> import os
            >>> import phanterpwa
            >>> from phanterpwa.configer import ProjectConfig
            >>> test_folder = os.path.join(os.path.dirname(phanterpwa.__file__), "tests", "test_configer_path", "project01")
            >>> cfg = ProjectConfig(test_folder)
            >>> cfg['PROJECT']['title']
            PhanterPWA
            >>> cfg['PROJECT']['title'] = 'PhanterPWA2'
            >>> cfg.save()
            >>> with open(os.path.join(test_folder, "config.json"), encoding="utf-8") as f:
            ...     content = json.load(f)
            >>> content['PROJECT']['title']
            PhanterPWA2
        """
        t = self._file
        with open(t, "w", encoding="utf-8") as f:
            json.dump(self._config, f, ensure_ascii=True, indent=4)

        if not os.path.exists(join(self._config["PROJECT"]["path"], 'project.ini')):
            for x in self.project_ini.sections():
                for y in self.project_ini.items(x):
                    v = str(self._config[x][y[0]])
                    self.project_ini[x][y[0]] = v
            # with open(join(self._config["PROJECT"]["path"], 'project.ini'), 'w', encoding="utf-8") as configfile:
            #     self._ini_project.write(configfile)

        if not os.path.exists(join(self._config["PROJECT"]["path"], 'secret.ini')):
            for x in self.project_secret_ini.sections():
                for y in self.project_secret_ini.items(x):
                    v = str(self._config[x][y[0]])
                    self.project_secret_ini[x][y[0]] = v
            # with open(join(self._config["PROJECT"]["path"], 'secret.ini'), 'w', encoding="utf-8") as configfile:
            #     self._ini_secret.write(configfile)

        for a in self.backend_ini:
            if not os.path.exists(join(self._config["PROJECT"]["path"], "backapps", a, 'app.ini')):
                for x in self.backend_ini[a].sections():
                    if x == "BACKEND":
                        for y in self.backend_ini[a].items(x):
                            v = str(self._config['BACKEND'][a][y[0]])
                            self.backend_ini[a][x][y[0]] = v
                # with open(join(self._config["PROJECT"]["path"], "backapps", a, 'app.ini'), 'w', encoding="utf-8") as configfile:
                #     self.backend_ini[a].write(configfile)

        for a in self.frontend_ini:
            if not os.path.exists(join(self._config["PROJECT"]["path"], "frontapps", a, 'app.ini')):
                for x in self.frontend_ini[a].sections():
                    if x == "FRONTEND":
                        for y in self.frontend_ini[a].items(x):
                            v = str(self._config['FRONTEND'][a][y[0]])
                            self.frontend_ini[a][x][y[0]] = v
                # with open(join(self._config["PROJECT"]["path"], "frontapps", a, 'app.ini'), 'w', encoding="utf-8") as configfile:
                #     self.frontend_ini[a].write(configfile)


    def __iter__(self):
        for c in self._config:
            yield c

    def __getitem__(self, i):
        return self._config[i]

    def __setitem__(self, i, v):
        self._config[i] = v
        self.save()

    def get(self, k, default=None):
        return self._config.get(k, default)
